from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType
from infrahub.core.query.branch import DeleteBranchRelationshipsQuery
from infrahub.log import get_logger

from ..shared import ArbitraryMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class DeletedBranchCleanupQuery(Query):
    """
    Find all unique edge branch names for which there is no Branch object
    """

    name = "deleted_branch_cleanup"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (b:Branch)
WITH collect(DISTINCT b.name) AS branch_names
MATCH ()-[e]->()
WHERE e.branch IS NOT NULL
AND NOT e.branch IN branch_names
RETURN DISTINCT (e.branch) AS branch_name
        """
        self.add_to_query(query)
        self.return_labels = ["branch_name"]


class DeleteOrphanRelationshipsQuery(Query):
    """
    Find all Relationship vertices that link to fewer than 2 Node vertices and delete them
    """

    name = "delete_orphan_relationships"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (r:Relationship)-[:IS_RELATED]-(n:Node)
WITH DISTINCT r, n
WITH r, count(*) AS node_count
WHERE node_count < 2
DETACH DELETE r
        """
        self.add_to_query(query)


class Migration032(ArbitraryMigration):
    """
    Delete edges for branches that were not completely deleted
    """

    name: str = "032_cleanup_deleted_branches"
    minimum_version: int = 31

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        migration_result = MigrationResult()

        try:
            log.info("Get partially deleted branch names...")
            orphaned_branches_query = await DeletedBranchCleanupQuery.init(db=db)
            await orphaned_branches_query.execute(db=db)

            orphaned_branch_names = []
            for result in orphaned_branches_query.get_results():
                branch_name = result.get_as_type("branch_name", str)
                orphaned_branch_names.append(branch_name)

            if not orphaned_branch_names:
                log.info("No partially deleted branches found. All done.")
                return migration_result

            log.info(f"Found {len(orphaned_branch_names)} orphaned branch names: {orphaned_branch_names}")

            for branch_name in orphaned_branch_names:
                log.info(f"Cleaning up branch '{branch_name}'...")
                delete_query = await DeleteBranchRelationshipsQuery.init(db=db, branch_name=branch_name)
                await delete_query.execute(db=db)
                log.info(f"Branch '{branch_name}' cleaned up.")

            log.info("Deleting orphaned relationships...")
            delete_relationships_query = await DeleteOrphanRelationshipsQuery.init(db=db)
            await delete_relationships_query.execute(db=db)
            log.info("Orphaned relationships deleted.")

        except Exception as exc:
            migration_result.errors.append(str(exc))
            log.exception("Error during branch cleanup")

        return migration_result



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from typing_extensions import Self

from infrahub.core.constants import NULL_VALUE, InfrahubKind, PermissionDecision
from infrahub.core.query import Query, QueryType
from infrahub.core.registry import registry
from infrahub.core.timestamp import Timestamp

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase
    from infrahub.permissions import AssignedPermissions


@dataclass
class GlobalPermission:
    action: str
    decision: int
    description: str = ""
    id: str = ""

    def __str__(self) -> str:
        decision = PermissionDecision(self.decision)
        return f"global:{self.action}:{decision.name.lower()}"

    @classmethod
    def from_string(cls, input: str) -> Self:
        parts = input.split(":")
        if len(parts) != 3 and parts[0] != "global":
            raise ValueError(f"{input} is not a valid format for a Global permission")

        return cls(action=parts[1], decision=PermissionDecision[parts[2].upper()])


@dataclass
class ObjectPermission:
    namespace: str
    name: str
    action: str
    decision: int
    description: str = ""
    id: str = ""

    def __str__(self) -> str:
        decision = PermissionDecision(self.decision)
        return f"object:{self.namespace}:{self.name}:{self.action}:{decision.name.lower()}"


class AccountGlobalPermissionQuery(Query):
    name: str = "account_global_permissions"
    type: QueryType = QueryType.READ

    def __init__(self, account_id: str, **kwargs: Any):
        self.account_id = account_id
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params["account_id"] = self.account_id

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic, is_isolated=False
        )
        self.params.update(branch_params)

        # ruff: noqa: E501
        query = """
        MATCH (account:%(generic_account_node)s)
        WHERE account.uuid = $account_id
        CALL (account) {
            MATCH (account)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN account as account1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH account, r1 as r
        WHERE r.status = "active"
        WITH account
        CALL (account) {
            MATCH (account)-[r1:IS_RELATED]->(:Relationship {name: "group_member"})<-[r2:IS_RELATED]-(account_group:%(account_group_node)s)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH account_group, r1, r2, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY account_group.uuid, r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            WITH account_group, head(collect(is_active)) as latest_is_active
            WHERE latest_is_active = TRUE
            RETURN account_group
        }
        WITH account_group

        CALL (account_group) {
            MATCH (account_group)-[r1:IS_RELATED]->(:Relationship {name: "role__accountgroups"})<-[r2:IS_RELATED]-(account_role:%(account_role_node)s)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH account_role, r1, r2, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY account_role.uuid, r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            WITH account_role, head(collect(is_active)) as latest_is_active
            WHERE latest_is_active = TRUE
            RETURN account_role
        }
        WITH account_role

        CALL (account_role) {
            MATCH (account_role)-[r1:IS_RELATED]->(:Relationship {name: "role__permissions"})<-[r2:IS_RELATED]-(global_permission:%(global_permission_node)s)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH global_permission, r1, r2, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY global_permission.uuid, r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            WITH global_permission, head(collect(is_active)) as latest_is_active
            WHERE latest_is_active = TRUE
            RETURN global_permission
        }
        WITH global_permission

        CALL (global_permission) {
            WITH global_permission
            MATCH (global_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "action"})-[r2:HAS_VALUE]->(global_permission_action:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN global_permission_action, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH global_permission, global_permission_action, is_active AS gpa_is_active
        WHERE gpa_is_active = TRUE

        CALL (global_permission) {
            MATCH (global_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "decision"})-[r2:HAS_VALUE]->(global_permission_decision:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN global_permission_decision, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH global_permission, global_permission_action, global_permission_decision, is_active AS gpd_is_active
        WHERE gpd_is_active = TRUE
        """ % {
            "branch_filter": branch_filter,
            "generic_account_node": InfrahubKind.GENERICACCOUNT,
            "account_group_node": InfrahubKind.ACCOUNTGROUP,
            "account_role_node": InfrahubKind.ACCOUNTROLE,
            "global_permission_node": InfrahubKind.GLOBALPERMISSION,
        }

        self.add_to_query(query)

        self.return_labels = ["global_permission", "global_permission_action", "global_permission_decision"]

    def get_permissions(self) -> list[GlobalPermission]:
        permissions: list[GlobalPermission] = []

        for result in self.get_results():
            permissions.append(
                GlobalPermission(
                    id=result.get("global_permission").get("uuid"),
                    action=result.get("global_permission_action").get("value"),
                    decision=result.get("global_permission_decision").get("value"),
                )
            )

        return permissions


class AccountObjectPermissionQuery(Query):
    name: str = "account_object_permissions"
    type: QueryType = QueryType.READ

    def __init__(self, account_id: str, **kwargs: Any):
        self.account_id = account_id
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params["account_id"] = self.account_id

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic, is_isolated=False
        )
        self.params.update(branch_params)

        query = """
        MATCH (account:%(generic_account_node)s)
        WHERE account.uuid = $account_id
        CALL (account) {
            MATCH (account)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN account as account1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH account, r1 as r
        WHERE r.status = "active"
        WITH account
        CALL (account) {
            MATCH (account)-[r1:IS_RELATED]->(:Relationship {name: "group_member"})<-[r2:IS_RELATED]-(account_group:%(account_group_node)s)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH account_group, r1, r2, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY account_group.uuid, r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            WITH account_group, head(collect(is_active)) as latest_is_active
            WHERE latest_is_active = TRUE
            RETURN account_group
        }
        WITH account_group

        CALL (account_group) {
            MATCH (account_group)-[r1:IS_RELATED]->(:Relationship {name: "role__accountgroups"})<-[r2:IS_RELATED]-(account_role:%(account_role_node)s)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH account_role, r1, r2, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY account_role.uuid, r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            WITH account_role, head(collect(is_active)) as latest_is_active
            WHERE latest_is_active = TRUE
            RETURN account_role
        }
        WITH account_role

        CALL (account_role) {
            MATCH (account_role)-[r1:IS_RELATED]->(:Relationship {name: "role__permissions"})<-[r2:IS_RELATED]-(object_permission:%(object_permission_node)s)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH object_permission, r1, r2, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY object_permission.uuid, r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            WITH object_permission, head(collect(is_active)) as latest_is_active
            WHERE latest_is_active = TRUE
            RETURN object_permission
        }
        WITH object_permission

        CALL (object_permission) {
            MATCH (object_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "namespace"})-[r2:HAS_VALUE]->(object_permission_namespace:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN object_permission_namespace, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH object_permission, object_permission_namespace, is_active AS opn_is_active
        WHERE opn_is_active = TRUE
        CALL (object_permission) {
            MATCH (object_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[r2:HAS_VALUE]->(object_permission_name:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN object_permission_name, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH object_permission, object_permission_namespace, object_permission_name, is_active AS opn_is_active
        WHERE opn_is_active = TRUE
        CALL (object_permission) {
            MATCH (object_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "action"})-[r2:HAS_VALUE]->(object_permission_action:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN object_permission_action, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH object_permission, object_permission_namespace, object_permission_name, object_permission_action, is_active AS opa_is_active
        WHERE opa_is_active = TRUE
        CALL (object_permission) {
            MATCH (object_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "decision"})-[r2:HAS_VALUE]->(object_permission_decision:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN object_permission_decision, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH object_permission, object_permission_namespace, object_permission_name, object_permission_action, object_permission_decision, is_active AS opd_is_active
        WHERE opd_is_active = TRUE
        """ % {
            "branch_filter": branch_filter,
            "account_group_node": InfrahubKind.ACCOUNTGROUP,
            "account_role_node": InfrahubKind.ACCOUNTROLE,
            "generic_account_node": InfrahubKind.GENERICACCOUNT,
            "object_permission_node": InfrahubKind.OBJECTPERMISSION,
        }

        self.add_to_query(query)

        self.return_labels = [
            "object_permission",
            "object_permission_namespace",
            "object_permission_name",
            "object_permission_action",
            "object_permission_decision",
        ]

    def get_permissions(self) -> list[ObjectPermission]:
        permissions: list[ObjectPermission] = []
        for result in self.get_results():
            permissions.append(
                ObjectPermission(
                    id=result.get("object_permission").get("uuid"),
                    namespace=result.get("object_permission_namespace").get("value"),
                    name=result.get("object_permission_name").get("value"),
                    action=result.get("object_permission_action").get("value"),
                    decision=result.get("object_permission_decision").get("value"),
                )
            )

        return permissions


async def fetch_permissions(account_id: str, db: InfrahubDatabase, branch: Branch) -> AssignedPermissions:
    query1 = await AccountGlobalPermissionQuery.init(db=db, branch=branch, account_id=account_id, branch_agnostic=True)
    await query1.execute(db=db)
    global_permissions = query1.get_permissions()

    query2 = await AccountObjectPermissionQuery.init(db=db, branch=branch, account_id=account_id)
    await query2.execute(db=db)
    object_permissions = query2.get_permissions()

    return {"global_permissions": global_permissions, "object_permissions": object_permissions}


class AccountRoleGlobalPermissionQuery(Query):
    name: str = "account_role_global_permissions"
    type: QueryType = QueryType.READ

    def __init__(self, role_id: str, **kwargs: Any):
        self.role_id = role_id
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params["role_id"] = self.role_id

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic, is_isolated=False
        )
        self.params.update(branch_params)

        # ruff: noqa: E501
        query = """
        MATCH (account_role:%(account_role_node)s)
        WHERE account_role.uuid = $role_id
        CALL (account_role) {
            MATCH (account_role)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN account_role as account_role1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH account_role, r1 as r
        WHERE r.status = "active"
        WITH account_role

        CALL (account_role) {
            MATCH (account_role)-[r1:IS_RELATED]->(:Relationship {name: "role__permissions"})<-[r2:IS_RELATED]-(global_permission:%(global_permission_node)s)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH global_permission, r1, r2, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY global_permission.uuid, r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            WITH global_permission, head(collect(is_active)) as latest_is_active
            WHERE latest_is_active = TRUE
            RETURN global_permission
        }
        WITH global_permission

        CALL (global_permission) {
            MATCH (global_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "action"})-[r2:HAS_VALUE]->(global_permission_action:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN global_permission_action, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH global_permission, global_permission_action, is_active AS gpa_is_active
        WHERE gpa_is_active = TRUE

        CALL (global_permission) {
            MATCH (global_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "decision"})-[r2:HAS_VALUE]->(global_permission_decision:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN global_permission_decision, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH global_permission, global_permission_action, global_permission_decision, is_active AS gpd_is_active
        WHERE gpd_is_active = TRUE
        """ % {
            "branch_filter": branch_filter,
            "account_role_node": InfrahubKind.ACCOUNTROLE,
            "global_permission_node": InfrahubKind.GLOBALPERMISSION,
        }

        self.add_to_query(query)

        self.return_labels = ["global_permission", "global_permission_action", "global_permission_decision"]

    def get_permissions(self) -> list[GlobalPermission]:
        permissions: list[GlobalPermission] = []

        for result in self.get_results():
            permissions.append(
                GlobalPermission(
                    id=result.get("global_permission").get("uuid"),
                    action=result.get("global_permission_action").get("value"),
                    decision=result.get("global_permission_decision").get("value"),
                )
            )

        return permissions


class AccountRoleObjectPermissionQuery(Query):
    name: str = "account_role_object_permissions"
    type: QueryType = QueryType.READ

    def __init__(self, role_id: str, **kwargs: Any):
        self.role_id = role_id
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params["role_id"] = self.role_id

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic, is_isolated=False
        )
        self.params.update(branch_params)

        query = """
        MATCH (account_role:%(account_role_node)s)
        WHERE account_role.uuid = $role_id
        CALL (account_role) {
            MATCH (account_role)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN account_role as account_role1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH account_role, r1 as r
        WHERE r.status = "active"
        WITH account_role

        CALL (account_role) {
            MATCH (account_role)-[r1:IS_RELATED]->(:Relationship {name: "role__permissions"})<-[r2:IS_RELATED]-(object_permission:%(object_permission_node)s)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH object_permission, r1, r2, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY object_permission.uuid, r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            WITH object_permission, head(collect(is_active)) as latest_is_active
            WHERE latest_is_active = TRUE
            RETURN object_permission
        }
        WITH object_permission

        CALL (object_permission) {
            MATCH (object_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "namespace"})-[r2:HAS_VALUE]->(object_permission_namespace:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN object_permission_namespace, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH object_permission, object_permission_namespace, is_active AS opn_is_active
        WHERE opn_is_active = TRUE
        CALL (object_permission) {
            MATCH (object_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[r2:HAS_VALUE]->(object_permission_name:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN object_permission_name, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH object_permission, object_permission_namespace, object_permission_name, is_active AS opn_is_active
        WHERE opn_is_active = TRUE
        CALL (object_permission) {
            MATCH (object_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "action"})-[r2:HAS_VALUE]->(object_permission_action:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN object_permission_action, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH object_permission, object_permission_namespace, object_permission_name, object_permission_action, is_active AS opa_is_active
        WHERE opa_is_active = TRUE
        CALL (object_permission) {
            MATCH (object_permission)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "decision"})-[r2:HAS_VALUE]->(object_permission_decision:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            RETURN object_permission_decision, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY r2.branch_level DESC, r2.from DESC, r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH object_permission, object_permission_namespace, object_permission_name, object_permission_action, object_permission_decision, is_active AS opd_is_active
        WHERE opd_is_active = TRUE
        """ % {
            "branch_filter": branch_filter,
            "account_role_node": InfrahubKind.ACCOUNTROLE,
            "object_permission_node": InfrahubKind.OBJECTPERMISSION,
        }

        self.add_to_query(query)

        self.return_labels = [
            "object_permission",
            "object_permission_namespace",
            "object_permission_name",
            "object_permission_action",
            "object_permission_decision",
        ]

    def get_permissions(self) -> list[ObjectPermission]:
        permissions: list[ObjectPermission] = []
        for result in self.get_results():
            permissions.append(
                ObjectPermission(
                    id=result.get("object_permission").get("uuid"),
                    namespace=result.get("object_permission_namespace").get("value"),
                    name=result.get("object_permission_name").get("value"),
                    action=result.get("object_permission_action").get("value"),
                    decision=result.get("object_permission_decision").get("value"),
                )
            )

        return permissions


async def fetch_role_permissions(role_id: str, db: InfrahubDatabase, branch: Branch) -> AssignedPermissions:
    query1 = await AccountRoleGlobalPermissionQuery.init(db=db, branch=branch, role_id=role_id, branch_agnostic=True)
    await query1.execute(db=db)
    global_permissions = query1.get_permissions()

    query2 = await AccountRoleObjectPermissionQuery.init(db=db, branch=branch, role_id=role_id)
    await query2.execute(db=db)
    object_permissions = query2.get_permissions()

    return {"global_permissions": global_permissions, "object_permissions": object_permissions}


class AccountTokenValidateQuery(Query):
    name: str = "account_token_validate"
    type: QueryType = QueryType.READ

    def __init__(self, token: str, **kwargs: Any):
        self.token = token
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic, is_isolated=False
        )
        self.params.update(branch_params)
        self.params.update(
            {
                "token_attr_name": "token",
                "token_relationship_name": "account__token",
                "token_value": self.token,
                "null_value": NULL_VALUE,
            }
        )

        query = """
// --------------
// get the active token node for this token value, if it exists
// --------------
MATCH (token_node:%(token_node_kind)s)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: $token_attr_name})
    -[r2:HAS_VALUE]->(av:AttributeValue { value: $token_value })
WHERE all(r in [r1, r2] WHERE (%(branch_filter)s))
ORDER BY r1.branch_level DESC, r1.from DESC, r1.status ASC, r2.branch_level DESC, r2.from DESC, r2.status ASC
LIMIT 1
WITH token_node
WHERE r1.status = "active" AND r2.status = "active"
// --------------
// get the expiration time
// --------------
OPTIONAL MATCH (token_node)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "expiration"})
    -[r2:HAS_VALUE]->(av)
WHERE all(r in [r1, r2] WHERE (%(branch_filter)s))
ORDER BY r1.branch_level DESC, r1.from DESC, r1.status ASC, r2.branch_level DESC, r2.from DESC, r2.status ASC
LIMIT 1
WITH token_node, CASE
    WHEN r1.status = "active" AND r2.status = "active" AND av.value <> $null_value THEN av.value
    ELSE NULL
END AS expiration
// --------------
// get the linked account node from the token node
// --------------
MATCH (token_node)-[r1:IS_RELATED]-(:Relationship {name: $token_relationship_name})-[r2:IS_RELATED]-(account_node:%(account_node_kind)s)
WHERE all(r in [r1, r2] WHERE (%(branch_filter)s))
ORDER BY r1.branch_level DESC, r1.from DESC, r1.status ASC, r2.branch_level DESC, r2.from DESC, r2.status ASC
LIMIT 1
WITH expiration, account_node
WHERE r1.status = "active" AND r2.status = "active"
        """ % {
            "branch_filter": branch_filter,
            "token_node_kind": InfrahubKind.ACCOUNTTOKEN,
            "account_node_kind": InfrahubKind.GENERICACCOUNT,
        }
        self.add_to_query(query)
        self.return_labels = ["account_node.uuid AS account_uuid", "expiration"]

    def get_account_id(self) -> str | None:
        """Return the account id that matched the query or a None."""
        result = self.get_result()
        if not result:
            return None
        account_uuid = result.get_as_str(label="account_uuid")
        expiration_with_tz = result.get_as_str(label="expiration")
        if expiration_with_tz is None:
            return account_uuid
        expiration = Timestamp(expiration_with_tz)
        if expiration < Timestamp():
            return None
        return account_uuid


async def validate_token(token: str, db: InfrahubDatabase, branch: Branch | str | None = None) -> str | None:
    branch = await registry.get_branch(db=db, branch=branch)
    query = await AccountTokenValidateQuery.init(db=db, branch=branch, token=token)
    await query.execute(db=db)
    return query.get_account_id()



from typing import Any, TypeVar, cast

from infrahub_sdk.protocols import CoreValidator

from infrahub.context import InfrahubContext
from infrahub.core.constants import ValidatorConclusion, ValidatorState
from infrahub.services import InfrahubServices

from .events import send_start_validator

ValidatorType = TypeVar("ValidatorType", bound=CoreValidator)


async def start_validator(
    service: InfrahubServices,
    validator: CoreValidator | None,
    validator_type: type[ValidatorType],
    proposed_change: str,
    context: InfrahubContext,
    data: dict[str, Any],
) -> ValidatorType:
    if validator:
        validator.conclusion.value = ValidatorConclusion.UNKNOWN.value
        validator.state.value = ValidatorState.QUEUED.value
        validator.started_at.value = ""
        validator.completed_at.value = ""
        await validator.save()
        validator = cast(ValidatorType, validator)
    else:
        data["proposed_change"] = proposed_change
        validator = await service.client.create(
            kind=validator_type,
            data=data,
        )
        await validator.save()

    await send_start_validator(
        service=service, validator=validator, proposed_change_id=proposed_change, context=context
    )

    return validator



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_transform = GenericSchema(
    name="Transformation",
    namespace="Core",
    description="Generic Transformation Object.",
    include_in_menu=False,
    icon="mdi:cog-transfer",
    label="Transformation",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["label__value"],
    branch=BranchSupportType.AWARE,
    documentation="/topics/proposed-change",
    uniqueness_constraints=[["name__value"]],
    attributes=[
        Attr(name="name", kind="Text", unique=True),
        Attr(name="label", kind="Text", optional=True),
        Attr(name="description", kind="Text", optional=True),
        Attr(name="timeout", kind="Number", default_value=10),
    ],
    relationships=[
        Rel(
            name="query",
            peer=InfrahubKind.GRAPHQLQUERY,
            identifier="graphql_query__transformation",
            kind=RelKind.ATTRIBUTE,
            cardinality=Cardinality.ONE,
            optional=False,
        ),
        Rel(
            name="repository",
            peer=InfrahubKind.GENERICREPOSITORY,
            kind=RelKind.ATTRIBUTE,
            cardinality=Cardinality.ONE,
            identifier="repository__transformation",
            optional=False,
        ),
        Rel(
            name="tags",
            peer=InfrahubKind.TAG,
            kind=RelKind.ATTRIBUTE,
            optional=True,
            cardinality=Cardinality.MANY,
        ),
    ],
)

core_transform_jinja2 = NodeSchema(
    name="TransformJinja2",
    namespace="Core",
    description="A file rendered from a Jinja2 template",
    include_in_menu=False,
    label="Transform Jinja2",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    inherit_from=[InfrahubKind.TRANSFORM],
    generate_profile=False,
    branch=BranchSupportType.AWARE,
    documentation="/topics/transformation",
    attributes=[
        Attr(name="template_path", kind="Text"),
    ],
)

core_transform_python = NodeSchema(
    name="TransformPython",
    namespace="Core",
    description="A transform function written in Python",
    include_in_menu=False,
    label="Transform Python",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    inherit_from=[InfrahubKind.TRANSFORM],
    generate_profile=False,
    branch=BranchSupportType.AWARE,
    documentation="/topics/transformation",
    attributes=[
        Attr(name="file_path", kind="Text"),
        Attr(name="class_name", kind="Text"),
        Attr(name="convert_query_response", kind="Boolean", optional=True, default_value=False),
    ],
)



import typer

from ..ctl import config


def load_configuration(value: str) -> str:
    """Load the configuration file using default environment variables or from the specified configuration file"""
    config.SETTINGS.load_and_exit(config_file=value)
    return value


CONFIG_PARAM = typer.Option(
    config.DEFAULT_CONFIG_FILE, "--config-file", envvar=config.ENVVAR_CONFIG_FILE, callback=load_configuration
)



from __future__ import annotations

from pathlib import Path
from typing import Optional

import typer
import yaml
from pydantic import ValidationError
from rich.console import Console
from rich.table import Table

from ..async_typer import AsyncTyper
from ..exceptions import FileNotValidError
from ..graphql import Mutation, Query
from ..schema.repository import InfrahubRepositoryConfig
from ..utils import read_file
from .client import initialize_client
from .parameters import CONFIG_PARAM
from .utils import init_logging

app = AsyncTyper()
console = Console()


def get_repository_config(repo_config_file: Path) -> InfrahubRepositoryConfig:
    try:
        config_file_data = load_repository_config_file(repo_config_file)
    except FileNotFoundError as exc:
        console.print(f"[red]File not found {exc}")
        raise typer.Exit(1) from exc
    except FileNotValidError as exc:
        console.print(f"[red]{exc.message}")
        raise typer.Exit(1) from exc

    try:
        data = InfrahubRepositoryConfig(**config_file_data)
    except ValidationError as exc:
        console.print(f"[red]Repository config file not valid, found {len(exc.errors())} error(s)")
        for error in exc.errors():
            loc_str = [str(item) for item in error["loc"]]
            console.print(f"  {'/'.join(loc_str)} | {error['msg']} ({error['type']})")
        raise typer.Exit(1) from exc

    return data


def load_repository_config_file(repo_config_file: Path) -> dict:
    yaml_data = read_file(file_path=repo_config_file)

    try:
        data = yaml.safe_load(yaml_data)
    except yaml.YAMLError as exc:
        raise FileNotValidError(name=str(repo_config_file)) from exc

    return data


@app.callback()
def callback() -> None:
    """
    Manage the repositories in a remote Infrahub instance.

    List, create, delete ..
    """


@app.command()
async def add(
    name: str,
    location: str,
    description: str = "",
    username: Optional[str] = None,
    password: str = "",
    ref: str = "",
    read_only: bool = False,
    debug: bool = False,
    _: str = CONFIG_PARAM,
) -> None:
    """Add a new repository."""

    init_logging(debug=debug)

    input_data = {
        "data": {
            "name": {"value": name},
            "location": {"value": location},
            "description": {"value": description},
        },
    }
    if read_only:
        input_data["data"]["ref"] = {"value": ref}
    else:
        input_data["data"]["default_branch"] = {"value": ref}

    client = initialize_client()

    if username or password:
        credential = await client.create(
            kind="CorePasswordCredential",
            name=name,
            username=username,
            password=password,
        )
        await credential.save(allow_upsert=True)
        input_data["data"]["credential"] = {"id": credential.id}

    query = Mutation(
        mutation="CoreReadOnlyRepositoryCreate" if read_only else "CoreRepositoryCreate",
        input_data=input_data,
        query={"ok": None},
    )

    await client.execute_graphql(query=query.render(), tracker="mutation-repository-create")


@app.command()
async def list(
    branch: Optional[str] = typer.Option(None, help="Branch on which to list repositories."),
    debug: bool = False,
    _: str = CONFIG_PARAM,
) -> None:
    init_logging(debug=debug)

    client = initialize_client()

    repo_status_query = {
        "CoreGenericRepository": {
            "edges": {
                "node": {
                    "__typename": None,
                    "name": {"value": None},
                    "operational_status": {"value": None},
                    "sync_status": {"value": None},
                    "internal_status": {"value": None},
                    "... on CoreReadOnlyRepository": {
                        "ref": {"value": None},
                    },
                }
            }
        },
    }

    query = Query(name="GetRepositoryStatus", query=repo_status_query)
    resp = await client.execute_graphql(query=query.render(), branch_name=branch, tracker="query-repository-list")

    table = Table(title="List of all Repositories")

    table.add_column("Name", justify="right", style="cyan", no_wrap=True)
    table.add_column("Type")
    table.add_column("Operational status")
    table.add_column("Sync status")
    table.add_column("Internal status")
    table.add_column("Ref")

    for repository_node in resp["CoreGenericRepository"]["edges"]:
        repository = repository_node["node"]

        table.add_row(
            repository["name"]["value"],
            repository["__typename"],
            repository["operational_status"]["value"],
            repository["sync_status"]["value"],
            repository["internal_status"]["value"],
            repository["ref"]["value"] if "ref" in repository else "",
        )

    console.print(table)



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from .constants import InfrahubClientMode
from .exceptions import NodeNotFoundError
from .utils import dict_hash

if TYPE_CHECKING:
    from .client import InfrahubClient, InfrahubClientSync
    from .node import InfrahubNode, InfrahubNodeSync, RelatedNodeBase
    from .schema import MainSchemaTypesAPI


class InfrahubGroupContextBase:
    """Base class for InfrahubGroupContext and InfrahubGroupContextSync"""

    def __init__(self) -> None:
        self.related_node_ids: list[str] = []
        self.related_group_ids: list[str] = []
        self.unused_member_ids: list[str] | None = None
        self.previous_members: Sequence[RelatedNodeBase] | None = None
        self.previous_children: list[RelatedNodeBase] | None = None
        self.identifier: str | None = None
        self.params: dict[str, str] = {}
        self.delete_unused_nodes: bool = False
        self.group_type: str = "CoreStandardGroup"
        self.group_params: dict[str, Any] = {}

    def set_properties(
        self,
        identifier: str,
        params: dict[str, str] | None = None,
        delete_unused_nodes: bool = False,
        group_type: str | None = None,
        group_params: dict[str, Any] | None = None,
        branch: str | None = None,
    ) -> None:
        """Setter method to set the values of identifier and params.

        Args:
            identifier: The new value for the identifier.
            params: A dictionary with new values for the params.
        """
        self.identifier = identifier
        self.params = params or {}
        self.delete_unused_nodes = delete_unused_nodes
        self.group_type = group_type or self.group_type
        self.group_params = group_params or {}
        self.branch = branch

    def _get_params_as_str(self) -> str:
        """Convert the params in dict format, into a string"""
        params_as_str: list[str] = []
        for key, value in self.params.items():
            params_as_str.append(f"{key}: {value!s}")
        return ", ".join(params_as_str)

    def _generate_group_name(self, suffix: str | None = None) -> str:
        group_name = self.identifier or "sdk"

        if suffix:
            group_name += f"-{suffix}"

        if self.params:
            group_name += f"-{dict_hash(self.params)}"

        return group_name

    def _generate_group_description(self, schema: MainSchemaTypesAPI) -> str:
        """Generate the description of the group from the params
        and ensure it's not longer than the maximum length of the description field."""
        if not self.params:
            return ""

        description_str = self._get_params_as_str()
        description = schema.get_attribute(name="description")
        if description and description.max_length and len(description_str) > description.max_length:
            length = description.max_length - 5
            return description_str[:length] + "..."

        return description_str


class InfrahubGroupContext(InfrahubGroupContextBase):
    """Represents a Infrahub GroupContext in an asynchronous context."""

    def __init__(self, client: InfrahubClient) -> None:
        super().__init__()
        self.client = client

    async def get_group(self, store_peers: bool = False) -> InfrahubNode | None:
        group_name = self._generate_group_name()
        try:
            group = await self.client.get(
                kind=self.group_type, name__value=group_name, include=["members"], branch=self.branch
            )
        except NodeNotFoundError:
            return None

        if not store_peers:
            return group

        self.previous_members = group._get_relationship_many(name="members").peers
        return group

    async def delete_unused(self) -> None:
        if self.previous_members and self.unused_member_ids:
            for member in self.previous_members:
                if member.id in self.unused_member_ids and member.typename:
                    await self.client.delete(kind=member.typename, id=member.id)

    async def add_related_nodes(self, ids: list[str], update_group_context: bool | None = None) -> None:
        """
        Add related Nodes IDs to the context.

        Args:
            ids (list[str]): List of node IDs to be added.
            update_group_context (Optional[bool], optional): Flag to control whether to update the group context.
        """
        if update_group_context is not False and (
            self.client.mode == InfrahubClientMode.TRACKING or self.client.update_group_context or update_group_context
        ):
            self.related_node_ids.extend(ids)

    async def add_related_groups(self, ids: list[str], update_group_context: bool | None = None) -> None:
        """
        Add related Groups IDs to the context.

        Args:
            ids (list[str]): List of group IDs to be added.
            update_group_context (Optional[bool], optional): Flag to control whether to update the group context.
        """
        if update_group_context is not False and (
            self.client.mode == InfrahubClientMode.TRACKING or self.client.update_group_context or update_group_context
        ):
            self.related_group_ids.extend(ids)

    async def update_group(self) -> None:
        """
        Create or update (using upsert) a CoreStandardGroup to store all the Nodes and Groups used during an execution.
        """
        members: list[str] = self.related_group_ids + self.related_node_ids

        if not members:
            return

        group_name = self._generate_group_name()
        schema = await self.client.schema.get(kind=self.group_type)
        description = self._generate_group_description(schema=schema)

        existing_group = None
        if self.delete_unused_nodes:
            existing_group = await self.get_group(store_peers=True)

        group = await self.client.create(
            kind=self.group_type,
            name=group_name,
            description=description,
            members=members,
            branch=self.branch,
            **self.group_params,
        )
        await group.save(allow_upsert=True, update_group_context=False)

        if not existing_group:
            return

        # Calculate how many nodes should be deleted
        self.unused_member_ids = set(existing_group.members.peer_ids) - set(members)  # type: ignore

        if not self.delete_unused_nodes:
            return

        await self.delete_unused()
        # TODO : create anoter "read" group. Could be based of the store items
        # Need to filters the store items inherited from CoreGroup to add them as children
        # Need to validate that it's UUIDas "key" if we want to implement other methods to store item


class InfrahubGroupContextSync(InfrahubGroupContextBase):
    """Represents a Infrahub GroupContext in an synchronous context."""

    def __init__(self, client: InfrahubClientSync) -> None:
        super().__init__()
        self.client = client

    def get_group(self, store_peers: bool = False) -> InfrahubNodeSync | None:
        group_name = self._generate_group_name()
        try:
            group = self.client.get(kind=self.group_type, name__value=group_name, include=["members"])
        except NodeNotFoundError:
            return None

        if not store_peers:
            return group

        self.previous_members = group._get_relationship_many(name="members").peers
        return group

    def delete_unused(self) -> None:
        if self.previous_members and self.unused_member_ids:
            for member in self.previous_members:
                if member.id in self.unused_member_ids and member.typename:
                    self.client.delete(kind=member.typename, id=member.id)

    def add_related_nodes(self, ids: list[str], update_group_context: bool | None = None) -> None:
        """
        Add related Nodes IDs to the context.

        Args:
            ids (list[str]): List of node IDs to be added.
            update_group_context (Optional[bool], optional): Flag to control whether to update the group context.
        """
        if update_group_context is not False and (
            self.client.mode == InfrahubClientMode.TRACKING or self.client.update_group_context or update_group_context
        ):
            self.related_node_ids.extend(ids)

    def add_related_groups(self, ids: list[str], update_group_context: bool | None = None) -> None:
        """
        Add related Groups IDs to the context.

        Args:
            ids (list[str]): List of group IDs to be added.
            update_group_context (Optional[bool], optional): Flag to control whether to update the group context.
        """
        if update_group_context is not False and (
            self.client.mode == InfrahubClientMode.TRACKING or self.client.update_group_context or update_group_context
        ):
            self.related_group_ids.extend(ids)

    def update_group(self) -> None:
        """
        Create or update (using upsert) a CoreStandardGroup to store all the Nodes and Groups used during an execution.
        """
        members: list[str] = self.related_node_ids + self.related_group_ids

        if not members:
            return

        group_name = self._generate_group_name()
        schema = self.client.schema.get(kind=self.group_type)
        description = self._generate_group_description(schema=schema)

        existing_group = None
        if self.delete_unused_nodes:
            existing_group = self.get_group(store_peers=True)

        group = self.client.create(
            kind=self.group_type,
            name=group_name,
            description=description,
            members=members,
            branch=self.branch,
            **self.group_params,
        )
        group.save(allow_upsert=True, update_group_context=False)

        if not existing_group:
            return

        # Calculate how many nodes should be deleted
        self.unused_member_ids = set(existing_group.members.peer_ids) - set(members)  # type: ignore

        if not self.delete_unused_nodes:
            return

        self.delete_unused()

        # TODO : create anoter "read" group. Could be based of the store items
        # Need to filters the store items inherited from CoreGroup to add them as children
        # Need to validate that it's UUIDas "key" if we want to implement other methods to store item



from graphene import Field, ObjectType, String
from graphene.types.generic import GenericScalar
from graphql import GraphQLResolveInfo

from infrahub.core import registry
from infrahub.core.convert_object_type.schema_mapping import get_schema_mapping


class FieldsMapping(ObjectType):
    mapping = GenericScalar(required=True)


async def fields_mapping_type_conversion_resolver(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
    source_kind: str,
    target_kind: str,
) -> dict:
    source_schema = registry.get_node_schema(name=source_kind, branch=info.context.branch)
    target_schema = registry.get_node_schema(name=target_kind, branch=info.context.branch)

    mapping = get_schema_mapping(source_schema=source_schema, target_schema=target_schema)
    mapping_dict = {field_name: model.model_dump(mode="json") for field_name, model in mapping.items()}
    return {"mapping": mapping_dict}


FieldsMappingTypeConversion = Field(
    FieldsMapping,
    source_kind=String(),
    target_kind=String(),
    description="Retrieve fields mapping for converting object type",
    resolver=fields_mapping_type_conversion_resolver,
    required=True,
)



from typing import Any

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_serializer

from infrahub.core.branch import Branch
from infrahub.core.path import SchemaPath
from infrahub.core.schema import GenericSchema, NodeSchema
from infrahub.core.schema.schema_branch import SchemaBranch


class SchemaConstraintValidatorRequest(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    branch: Branch = Field(..., description="The name of the branch to target")
    constraint_name: str = Field(..., description="The name of the constraint to validate")
    node_schema: NodeSchema | GenericSchema = Field(..., description="Schema of Node or Generic to validate")
    schema_path: SchemaPath = Field(..., description="SchemaPath to the element of the schema to validate")
    schema_branch: SchemaBranch = Field(..., description="SchemaBranch of the element to validate")

    @model_serializer()
    def serialize_model(self) -> dict[str, Any]:
        return {
            "branch": self.branch.model_dump(),
            "constraint_name": self.constraint_name,
            "node_schema": self.node_schema.model_dump(),
            "schema_path": self.schema_path.model_dump(),
            "schema_branch": self.schema_branch.to_dict_schema_object(),
        }

    @field_validator("schema_branch", mode="before")
    @classmethod
    def validate_schema_branch(cls, value: Any) -> SchemaBranch:
        return SchemaBranch.validate(data=value)


class SchemaViolation(BaseModel):
    node_id: str
    node_kind: str
    display_label: str
    full_display_label: str
    message: str = ""



from __future__ import annotations

import importlib
import typing
from datetime import datetime
from typing import TYPE_CHECKING

import graphene
from graphene.types.generic import GenericScalar
from pydantic import EmailStr, HttpUrl, IPvAnyAddress, Json

from infrahub.core import registry

if TYPE_CHECKING:
    from infrahub.core.attribute import BaseAttribute
    from infrahub.graphql.mutations.attribute import BaseAttributeCreate, BaseAttributeUpdate
    from infrahub.graphql.types.attribute import BaseAttribute as BaseAttributeType

DEFAULT_MODULE_ATTRIBUTE = "infrahub.core.attribute"
DEFAULT_MODULE_GRAPHQL_INPUT = "infrahub.graphql.mutations.attribute"
DEFAULT_MODULE_GRAPHQL_QUERY = "infrahub.graphql.types"


class InfrahubDataType:
    label: str
    graphql_query: str
    graphql_create: str
    graphql_update: str
    graphql_filter: type
    graphql: type
    infrahub: str
    pydantic: type

    @classmethod
    def __init_subclass__(cls, **kwargs: typing.Any) -> None:
        super().__init_subclass__(**kwargs)
        registry.data_type[cls.label] = cls

    def __str__(self) -> str:
        return self.label

    @classmethod
    def get_infrahub_class(cls) -> type[BaseAttribute]:
        if not isinstance(cls.infrahub, str):
            return cls.infrahub
        module = importlib.import_module(DEFAULT_MODULE_ATTRIBUTE)
        return getattr(module, cls.infrahub)

    @classmethod
    def get_graphql_create(cls) -> type[BaseAttributeCreate]:
        if not isinstance(cls.graphql_create, str):
            return cls.graphql_create
        module = importlib.import_module(DEFAULT_MODULE_GRAPHQL_INPUT)
        return getattr(module, cls.graphql_create)

    @classmethod
    def get_graphql_update(cls) -> type[BaseAttributeUpdate]:
        if not isinstance(cls.graphql_update, str):
            return cls.graphql_update
        module = importlib.import_module(DEFAULT_MODULE_GRAPHQL_INPUT)
        return getattr(module, cls.graphql_update)

    @classmethod
    def get_graphql_type(cls) -> type[BaseAttributeType]:
        if not isinstance(cls.graphql_query, str):
            return cls.graphql_query
        module = importlib.import_module(DEFAULT_MODULE_GRAPHQL_QUERY)
        return getattr(module, cls.graphql_query)

    @classmethod
    def get_graphql_type_name(cls) -> str:
        return cls.get_graphql_type().__name__

    @classmethod
    def get_graphql_filters(
        cls, name: str, include_properties: bool = True, include_isnull: bool = False
    ) -> dict[str, typing.Any]:
        filters: dict[str, typing.Any] = {}
        attr_class = cls.get_infrahub_class()
        filters[f"{name}__value"] = cls.graphql_filter()
        filters[f"{name}__values"] = graphene.List(cls.graphql_filter)
        if include_isnull:
            filters[f"{name}__isnull"] = graphene.Boolean()

        if not include_properties:
            return filters

        for node_prop in attr_class._node_properties:
            filters[f"{name}__{node_prop}__id"] = graphene.ID()

        for flag_prop in attr_class._flag_properties:
            filters[f"{name}__{flag_prop}"] = graphene.Boolean()

        return filters


class Default(InfrahubDataType):
    label: str = "Default"
    graphql = graphene.String
    graphql_query = "BaseAttribute"
    graphql_create = "BaseAttributeCreate"
    graphql_update = "BaseAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "BaseAttribute"


class ID(InfrahubDataType):
    label: str = "ID"
    graphql = graphene.ID
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "String"


class Text(InfrahubDataType):
    label: str = "Text"
    graphql = graphene.String
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "String"


class TextArea(Text):
    label: str = "TextArea"
    graphql = graphene.String
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "String"


class DateTime(InfrahubDataType):
    label: str = "DateTime"
    graphql = graphene.DateTime
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.DateTime
    infrahub = "DateTime"


class Email(InfrahubDataType):
    label: str = "Email"
    graphql = graphene.String
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "String"


class Password(InfrahubDataType):
    label: str = "Password"
    graphql = graphene.String
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "String"
    pydantic = str


class HashedPassword(InfrahubDataType):
    label: str = "Password"
    graphql = graphene.String
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "HashedPassword"


class URL(InfrahubDataType):
    label: str = "URL"
    graphql = graphene.String
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "URL"


class File(InfrahubDataType):
    label: str = "File"
    graphql = graphene.String
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "String"


class MacAddress(InfrahubDataType):
    label: str = "MacAddress"
    graphql = graphene.String
    graphql_query = "MacAddressType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "MacAddress"


class Color(InfrahubDataType):
    label: str = "Color"
    graphql = graphene.String
    graphql_query = "TextAttributeType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "String"


class Dropdown(InfrahubDataType):
    label: str = "Dropdown"
    graphql = graphene.String
    graphql_query = "DropdownType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "Dropdown"


class Number(InfrahubDataType):
    label: str = "Number"
    graphql = graphene.BigInt
    graphql_query = "NumberAttributeType"
    graphql_create = "NumberAttributeCreate"
    graphql_update = "NumberAttributeUpdate"
    graphql_filter = graphene.BigInt
    infrahub = "Integer"


class NumberPool(Number):
    label: str = "Number Pool"


class Bandwidth(InfrahubDataType):
    label: str = "Bandwidth"
    graphql = graphene.Int
    graphql_query = "NumberAttributeType"
    graphql_create = "NumberAttributeCreate"
    graphql_update = "NumberAttributeUpdate"
    graphql_filter = graphene.Int
    infrahub = "Integer"


class IPHost(InfrahubDataType):
    label: str = "IPHost"
    graphql = graphene.String
    graphql_query = "IPHostType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "IPHost"


class IPNetwork(InfrahubDataType):
    label: str = "IPNetwork"
    graphql = graphene.String
    graphql_query = "IPNetworkType"
    graphql_create = "TextAttributeCreate"
    graphql_update = "TextAttributeUpdate"
    graphql_filter = graphene.String
    infrahub = "IPNetwork"


class Boolean(InfrahubDataType):
    label: str = "Boolean"
    graphql = graphene.Boolean
    graphql_query = "CheckboxAttributeType"
    graphql_create = "CheckboxAttributeCreate"
    graphql_update = "CheckboxAttributeUpdate"
    graphql_filter = graphene.Boolean
    infrahub = "Boolean"


class Checkbox(InfrahubDataType):
    label: str = "Checkbox"
    graphql = graphene.Boolean
    graphql_query = "CheckboxAttributeType"
    graphql_create = "CheckboxAttributeCreate"
    graphql_update = "CheckboxAttributeUpdate"
    graphql_filter = graphene.Boolean
    infrahub = "Boolean"


class List(InfrahubDataType):
    label: str = "List"
    graphql = GenericScalar
    graphql_query = "ListAttributeType"
    graphql_create = "ListAttributeCreate"
    graphql_update = "ListAttributeUpdate"
    graphql_filter = GenericScalar
    infrahub = "ListAttribute"


class JSON(InfrahubDataType):
    label: str = "JSON"
    graphql = GenericScalar
    graphql_query = "JSONAttributeType"
    graphql_create = "JSONAttributeCreate"
    graphql_update = "JSONAttributeUpdate"
    graphql_filter = GenericScalar
    infrahub = "JSONAttribute"


class Any(InfrahubDataType):
    label: str = "Any"
    graphql = GenericScalar
    graphql_query = "AnyAttributeType"
    graphql_create = "AnyAttributeCreate"
    graphql_update = "AnyAttributeUpdate"
    graphql_filter = GenericScalar
    infrahub = "AnyAttribute"


ATTRIBUTE_TYPES: dict[str, type[InfrahubDataType]] = {
    "ID": ID,
    "Dropdown": Dropdown,
    "Text": Text,
    "TextArea": TextArea,
    "DateTime": DateTime,
    "Email": Email,
    "Password": Password,
    "HashedPassword": HashedPassword,
    "URL": URL,
    "File": File,
    "MacAddress": MacAddress,
    "Color": Color,
    "Number": Number,
    "NumberPool": NumberPool,
    "Bandwidth": Bandwidth,
    "IPHost": IPHost,
    "IPNetwork": IPNetwork,
    "Boolean": Boolean,
    "Checkbox": Checkbox,
    "List": List,
    "JSON": JSON,
    "Any": Any,
}

ATTRIBUTE_PYTHON_TYPES: dict[str, type] = {
    "ID": int,  # Assuming IDs are integers
    "Dropdown": str,  # Dropdowns can be represented as strings
    "Text": str,
    "TextArea": str,
    "DateTime": datetime,
    "Email": EmailStr,
    "Password": str,  # Passwords can be any string
    "HashedPassword": str,  # Hashed passwords are also strings
    "URL": HttpUrl,
    "File": str,  # File paths or identifiers as strings
    "MacAddress": str,  # MAC addresses can be straightforward strings
    "Color": str,  # Colors often represented as hex strings
    "Number": float,  # Numbers can be floats for general use
    "NumberPool": float,  # Numbers can be floats for general use
    "Bandwidth": float,  # Bandwidth in some units, represented as a float
    "IPHost": IPvAnyAddress,  # type: ignore[dict-item]
    "IPNetwork": str,
    "Boolean": bool,
    "Checkbox": bool,  # Checkboxes represent boolean values
    "List": list[Any],  # Lists can contain any type of items
    "JSON": Json,  # Pydantic's Json type handles arbitrary JSON objects
    "Any": Any,  # Any type allowed
}

ATTRIBUTE_KIND_LABELS = list(ATTRIBUTE_TYPES.keys())

# Data types supporting large values, which can therefore not be indexed in neo4j.
LARGE_ATTRIBUTE_TYPES = [TextArea, JSON, List]


def get_attribute_type(kind: str = "Default") -> type[InfrahubDataType]:
    """Return an InfrahubDataType object for a given kind
    If no kind is provided, return the default one."""
    return ATTRIBUTE_TYPES.get(kind, Default)



from neo4j.graph import Node as Neo4jNode
from neo4j.graph import Path as Neo4jPath

from infrahub.core.constants import DiffAction, RelationshipCardinality
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.query import QueryResult
from infrahub.core.timestamp import Timestamp

from ..model.path import (
    ConflictSelection,
    EnrichedDiffAttribute,
    EnrichedDiffConflict,
    EnrichedDiffNode,
    EnrichedDiffProperty,
    EnrichedDiffRelationship,
    EnrichedDiffRoot,
    EnrichedDiffRootMetadata,
    EnrichedDiffSingleRelationship,
    NodeIdentifier,
    deserialize_tracking_id,
)
from ..parent_node_adder import DiffParentNodeAdder, ParentNodeAddRequest


class EnrichedDiffDeserializer:
    def __init__(self, parent_adder: DiffParentNodeAdder) -> None:
        self.parent_adder = parent_adder
        self._diff_root_map: dict[str, EnrichedDiffRoot] = {}
        self._diff_node_map: dict[tuple[str, NodeIdentifier], EnrichedDiffNode] = {}
        self._diff_node_attr_map: dict[tuple[str, NodeIdentifier, str], EnrichedDiffAttribute] = {}
        self._diff_node_rel_group_map: dict[tuple[str, NodeIdentifier, str], EnrichedDiffRelationship] = {}
        self._diff_node_rel_element_map: dict[tuple[str, NodeIdentifier, str, str], EnrichedDiffSingleRelationship] = {}
        self._diff_prop_map: dict[
            tuple[str, NodeIdentifier, str, str] | tuple[str, str, str, str, str], EnrichedDiffProperty
        ] = {}
        # {EnrichedDiffRoot: [(NodeIdentifier, parents_path: Neo4jPath), ...]}
        self._parents_path_map: dict[EnrichedDiffRoot, list[tuple[NodeIdentifier, Neo4jPath]]] = {}

    def initialize(self) -> None:
        self._diff_root_map = {}
        self._diff_node_map = {}
        self._diff_node_attr_map = {}
        self._diff_node_rel_group_map = {}
        self._diff_node_rel_element_map = {}
        self._diff_prop_map = {}
        self._parents_path_map = {}

    def _track_parents_path(
        self, enriched_root: EnrichedDiffRoot, node_identifier: NodeIdentifier, parents_path: Neo4jPath
    ) -> None:
        if enriched_root not in self._parents_path_map:
            self._parents_path_map[enriched_root] = []
        self._parents_path_map[enriched_root].append((node_identifier, parents_path))

    async def read_result(self, result: QueryResult, include_parents: bool) -> None:
        enriched_root = self._deserialize_diff_root(root_node=result.get_node("diff_root"))
        node_node = result.get(label="diff_node")
        if not isinstance(node_node, Neo4jNode):
            return
        enriched_node = self._deserialize_diff_node(node_node=node_node, enriched_root=enriched_root)

        if include_parents:
            parents_path = result.get("parents_path")
            if parents_path and isinstance(parents_path, Neo4jPath):
                self._track_parents_path(
                    enriched_root=enriched_root, node_identifier=enriched_node.identifier, parents_path=parents_path
                )

        node_conflict_node = result.get(label="diff_node_conflict")
        if isinstance(node_conflict_node, Neo4jNode) and not enriched_node.conflict:
            conflict = self.deserialize_conflict(diff_conflict_node=node_conflict_node)
            enriched_node.conflict = conflict
        self._deserialize_attributes(result=result, enriched_root=enriched_root, enriched_node=enriched_node)
        self._deserialize_relationships(result=result, enriched_root=enriched_root, enriched_node=enriched_node)

    async def deserialize(self, include_parents: bool = True) -> list[EnrichedDiffRoot]:
        if include_parents:
            self._deserialize_parents()

        return list(self._diff_root_map.values())

    def _deserialize_attributes(
        self, result: QueryResult, enriched_root: EnrichedDiffRoot, enriched_node: EnrichedDiffNode
    ) -> None:
        for attribute_result in result.get_nested_node_collection("diff_attributes"):
            diff_attr_node, diff_attr_property_node, diff_attr_property_conflict = attribute_result
            if diff_attr_node is None:
                continue
            enriched_attribute = self._deserialize_diff_attr(
                diff_attr_node=diff_attr_node, enriched_root=enriched_root, enriched_node=enriched_node
            )
            if diff_attr_property_node is None:
                continue
            enriched_property = self._deserialize_diff_attr_property(
                diff_attr_property_node=diff_attr_property_node,
                enriched_attr=enriched_attribute,
                enriched_node=enriched_node,
                enriched_root=enriched_root,
            )
            if diff_attr_property_conflict:
                conflict = self.deserialize_conflict(diff_conflict_node=diff_attr_property_conflict)
                enriched_property.conflict = conflict

    def _deserialize_relationships(
        self, result: QueryResult, enriched_root: EnrichedDiffRoot, enriched_node: EnrichedDiffNode
    ) -> None:
        for relationship_result in result.get_nested_node_collection("diff_relationships"):
            group_node, element_node, element_conflict, property_node, property_conflict = relationship_result
            enriched_relationship_group = None
            if group_node:
                enriched_relationship_group = self._deserialize_diff_relationship_group(
                    relationship_group_node=group_node, enriched_root=enriched_root, enriched_node=enriched_node
                )
            if element_node is None or property_node is None or enriched_relationship_group is None:
                continue

            enriched_relationship_element = self._deserialize_diff_relationship_element(
                relationship_element_node=element_node,
                enriched_relationship_group=enriched_relationship_group,
                enriched_node=enriched_node,
                enriched_root=enriched_root,
            )
            if element_conflict and not enriched_relationship_element.conflict:
                conflict = self.deserialize_conflict(diff_conflict_node=element_conflict)
                enriched_relationship_element.conflict = conflict
            element_property = self._deserialize_diff_relationship_element_property(
                relationship_element_property_node=property_node,
                enriched_relationship_element=enriched_relationship_element,
                enriched_relationship_group=enriched_relationship_group,
                enriched_node=enriched_node,
                enriched_root=enriched_root,
            )
            if property_conflict:
                conflict = self.deserialize_conflict(diff_conflict_node=property_conflict)
                element_property.conflict = conflict

    def _deserialize_parents(self) -> None:
        for enriched_root, node_path_tuples in self._parents_path_map.items():
            self.parent_adder.initialize(enriched_diff_root=enriched_root)
            for node_identifier, parents_path in node_path_tuples:
                # Remove the node itself from the path
                parents_path_slice = parents_path.nodes[1:]

                # TODO Ensure the list is even
                current_node_identifier = node_identifier
                for rel, parent in zip(parents_path_slice[::2], parents_path_slice[1::2], strict=False):
                    parent_identifier = NodeIdentifier(
                        uuid=parent.get("uuid"),
                        kind=parent.get("kind"),
                        db_id=parent.get("db_id"),
                    )
                    parent_request = ParentNodeAddRequest(
                        node_identifier=current_node_identifier,
                        parent_identifier=parent_identifier,
                        parent_label=parent.get("label"),
                        parent_rel_name=rel.get("name"),
                        parent_rel_identifier=rel.get("identifier"),
                        parent_rel_cardinality=RelationshipCardinality(rel.get("cardinality")),
                        parent_rel_label=rel.get("label"),
                    )
                    self.parent_adder.add_parent(parent_request=parent_request)
                    current_node_identifier = parent_identifier

    @classmethod
    def _get_str_or_none_property_value(cls, node: Neo4jNode, property_name: str) -> str | None:
        value_raw = node.get(property_name)
        return str(value_raw) if value_raw is not None else None

    def _deserialize_diff_root(self, root_node: Neo4jNode) -> EnrichedDiffRoot:
        root_uuid = str(root_node.get("uuid"))
        if root_uuid in self._diff_root_map:
            return self._diff_root_map[root_uuid]
        root_empty = self.build_diff_root_metadata(root_node=root_node)
        enriched_root = EnrichedDiffRoot.from_root_metadata(empty_root=root_empty)
        self._diff_root_map[root_uuid] = enriched_root
        return enriched_root

    @classmethod
    def build_diff_root_metadata(cls, root_node: Neo4jNode) -> EnrichedDiffRootMetadata:
        from_time = Timestamp(str(root_node.get("from_time")))
        to_time = Timestamp(str(root_node.get("to_time")))
        partner_uuid = cls._get_str_or_none_property_value(node=root_node, property_name="partner_uuid")
        tracking_id_str = str(root_node.get("tracking_id"))
        tracking_id = deserialize_tracking_id(tracking_id_str=tracking_id_str)
        return EnrichedDiffRootMetadata(
            base_branch_name=str(root_node.get("base_branch")),
            diff_branch_name=str(root_node.get("diff_branch")),
            from_time=from_time,
            to_time=to_time,
            uuid=str(root_node.get("uuid")),
            partner_uuid=partner_uuid,
            tracking_id=tracking_id,
            num_added=int(root_node.get("num_added", 0)),
            num_updated=int(root_node.get("num_updated", 0)),
            num_removed=int(root_node.get("num_removed", 0)),
            num_conflicts=int(root_node.get("num_conflicts", 0)),
            contains_conflict=str(root_node.get("contains_conflict")).lower() == "true",
            exists_on_database=True,
        )

    def _deserialize_diff_node(self, node_node: Neo4jNode, enriched_root: EnrichedDiffRoot) -> EnrichedDiffNode:
        node_uuid = str(node_node.get("uuid"))
        node_kind = str(node_node.get("kind"))
        node_db_id = node_node.get("db_id")
        node_identifier = NodeIdentifier(uuid=node_uuid, kind=node_kind, db_id=node_db_id)
        node_key = (enriched_root.uuid, node_identifier)
        if node_key in self._diff_node_map:
            return self._diff_node_map[node_key]

        timestamp_str = self._get_str_or_none_property_value(node=node_node, property_name="changed_at")
        enriched_node = EnrichedDiffNode(
            identifier=node_identifier,
            label=str(node_node.get("label")),
            changed_at=Timestamp(timestamp_str) if timestamp_str else None,
            action=DiffAction(str(node_node.get("action"))),
            is_node_kind_migration=bool(node_node.get("is_node_kind_migration")),
            path_identifier=str(node_node.get("path_identifier")),
            num_added=int(node_node.get("num_added", 0)),
            num_updated=int(node_node.get("num_updated", 0)),
            num_removed=int(node_node.get("num_removed", 0)),
            num_conflicts=int(node_node.get("num_conflicts", 0)),
            contains_conflict=str(node_node.get("contains_conflict")).lower() == "true",
        )
        self._diff_node_map[node_key] = enriched_node
        enriched_root.nodes.add(enriched_node)
        return enriched_node

    def _deserialize_diff_attr(
        self, diff_attr_node: Neo4jNode, enriched_root: EnrichedDiffRoot, enriched_node: EnrichedDiffNode
    ) -> EnrichedDiffAttribute:
        attr_name = str(diff_attr_node.get("name"))
        attr_key = (enriched_root.uuid, enriched_node.identifier, attr_name)
        if attr_key in self._diff_node_attr_map:
            return self._diff_node_attr_map[attr_key]

        enriched_attr = EnrichedDiffAttribute(
            name=str(diff_attr_node.get("name")),
            changed_at=Timestamp(str(diff_attr_node.get("changed_at"))),
            path_identifier=str(diff_attr_node.get("path_identifier")),
            action=DiffAction(str(diff_attr_node.get("action"))),
            num_added=int(diff_attr_node.get("num_added", 0)),
            num_updated=int(diff_attr_node.get("num_updated", 0)),
            num_removed=int(diff_attr_node.get("num_removed", 0)),
            num_conflicts=int(diff_attr_node.get("num_conflicts", 0)),
            contains_conflict=str(diff_attr_node.get("contains_conflict")).lower() == "true",
        )
        self._diff_node_attr_map[attr_key] = enriched_attr
        enriched_node.attributes.add(enriched_attr)
        return enriched_attr

    def _deserialize_diff_relationship_group(
        self, relationship_group_node: Neo4jNode, enriched_root: EnrichedDiffRoot, enriched_node: EnrichedDiffNode
    ) -> EnrichedDiffRelationship:
        diff_rel_name = str(relationship_group_node.get("name"))
        rel_key = (enriched_root.uuid, enriched_node.identifier, diff_rel_name)
        if rel_key in self._diff_node_rel_group_map:
            return self._diff_node_rel_group_map[rel_key]

        timestamp_str = relationship_group_node.get("changed_at")
        enriched_relationship = EnrichedDiffRelationship(
            name=relationship_group_node.get("name"),
            identifier=relationship_group_node.get("identifier"),
            label=relationship_group_node.get("label"),
            cardinality=RelationshipCardinality(relationship_group_node.get("cardinality")),
            changed_at=Timestamp(timestamp_str) if timestamp_str else None,
            action=DiffAction(str(relationship_group_node.get("action"))),
            path_identifier=str(relationship_group_node.get("path_identifier")),
            num_added=int(relationship_group_node.get("num_added", 0)),
            num_conflicts=int(relationship_group_node.get("num_conflicts", 0)),
            num_removed=int(relationship_group_node.get("num_removed", 0)),
            num_updated=int(relationship_group_node.get("num_updated", 0)),
            contains_conflict=str(relationship_group_node.get("contains_conflict")).lower() == "true",
        )

        self._diff_node_rel_group_map[rel_key] = enriched_relationship
        enriched_node.relationships.add(enriched_relationship)
        return enriched_relationship

    def _deserialize_diff_relationship_element(
        self,
        relationship_element_node: Neo4jNode,
        enriched_relationship_group: EnrichedDiffRelationship,
        enriched_node: EnrichedDiffNode,
        enriched_root: EnrichedDiffRoot,
    ) -> EnrichedDiffSingleRelationship:
        diff_element_peer_id = str(relationship_element_node.get("peer_id"))
        rel_element_key = (
            enriched_root.uuid,
            enriched_node.identifier,
            enriched_relationship_group.name,
            diff_element_peer_id,
        )
        if rel_element_key in self._diff_node_rel_element_map:
            return self._diff_node_rel_element_map[rel_element_key]

        peer_label = self._get_str_or_none_property_value(node=relationship_element_node, property_name="peer_label")
        enriched_rel_element = EnrichedDiffSingleRelationship(
            changed_at=Timestamp(str(relationship_element_node.get("changed_at"))),
            action=DiffAction(str(relationship_element_node.get("action"))),
            peer_id=diff_element_peer_id,
            peer_label=peer_label,
            path_identifier=str(relationship_element_node.get("path_identifier")),
            num_added=int(relationship_element_node.get("num_added", 0)),
            num_updated=int(relationship_element_node.get("num_updated", 0)),
            num_removed=int(relationship_element_node.get("num_removed", 0)),
            num_conflicts=int(relationship_element_node.get("num_conflicts", 0)),
            contains_conflict=str(relationship_element_node.get("contains_conflict")).lower() == "true",
        )
        enriched_relationship_group.relationships.add(enriched_rel_element)
        self._diff_node_rel_element_map[rel_element_key] = enriched_rel_element
        return enriched_rel_element

    def _property_node_to_enriched_property(self, property_node: Neo4jNode) -> EnrichedDiffProperty:
        previous_value = self._get_str_or_none_property_value(node=property_node, property_name="previous_value")
        new_value = self._get_str_or_none_property_value(node=property_node, property_name="new_value")
        previous_label = self._get_str_or_none_property_value(node=property_node, property_name="previous_label")
        new_label = self._get_str_or_none_property_value(node=property_node, property_name="new_label")
        return EnrichedDiffProperty(
            property_type=DatabaseEdgeType(str(property_node.get("property_type"))),
            changed_at=Timestamp(str(property_node.get("changed_at"))),
            previous_value=previous_value,
            new_value=new_value,
            previous_label=previous_label,
            new_label=new_label,
            action=DiffAction(str(property_node.get("action"))),
            path_identifier=str(property_node.get("path_identifier")),
        )

    def _deserialize_diff_attr_property(
        self,
        diff_attr_property_node: Neo4jNode,
        enriched_attr: EnrichedDiffAttribute,
        enriched_node: EnrichedDiffNode,
        enriched_root: EnrichedDiffRoot,
    ) -> EnrichedDiffProperty:
        diff_prop_type = str(diff_attr_property_node.get("property_type"))
        attr_property_key = (enriched_root.uuid, enriched_node.identifier, enriched_attr.name, diff_prop_type)
        if attr_property_key in self._diff_prop_map:
            return self._diff_prop_map[attr_property_key]

        enriched_property = self._property_node_to_enriched_property(property_node=diff_attr_property_node)
        enriched_attr.properties.add(enriched_property)
        self._diff_prop_map[attr_property_key] = enriched_property
        return enriched_property

    def _deserialize_diff_relationship_element_property(
        self,
        relationship_element_property_node: Neo4jNode,
        enriched_relationship_element: EnrichedDiffSingleRelationship,
        enriched_relationship_group: EnrichedDiffRelationship,
        enriched_node: EnrichedDiffNode,
        enriched_root: EnrichedDiffRoot,
    ) -> EnrichedDiffProperty:
        diff_prop_type = str(relationship_element_property_node.get("property_type"))
        rel_property_key = (
            enriched_root.uuid,
            enriched_node.uuid,
            enriched_relationship_group.name,
            enriched_relationship_element.peer_id,
            diff_prop_type,
        )
        if rel_property_key in self._diff_prop_map:
            return self._diff_prop_map[rel_property_key]

        enriched_property = self._property_node_to_enriched_property(property_node=relationship_element_property_node)
        self._diff_prop_map[rel_property_key] = enriched_property
        enriched_relationship_element.properties.add(enriched_property)
        return enriched_property

    def deserialize_conflict(self, diff_conflict_node: Neo4jNode) -> EnrichedDiffConflict:
        base_branch_value = self._get_str_or_none_property_value(
            node=diff_conflict_node, property_name="base_branch_value"
        )
        diff_branch_value = self._get_str_or_none_property_value(
            node=diff_conflict_node, property_name="diff_branch_value"
        )
        base_branch_label = self._get_str_or_none_property_value(
            node=diff_conflict_node, property_name="base_branch_label"
        )
        diff_branch_label = self._get_str_or_none_property_value(
            node=diff_conflict_node, property_name="diff_branch_label"
        )
        base_timestamp_str = self._get_str_or_none_property_value(
            node=diff_conflict_node, property_name="base_branch_changed_at"
        )
        diff_timestamp_str = self._get_str_or_none_property_value(
            node=diff_conflict_node, property_name="diff_branch_changed_at"
        )
        selected_branch = self._get_str_or_none_property_value(node=diff_conflict_node, property_name="selected_branch")
        resolvable = str(diff_conflict_node.get("resolvable")).lower() == "true"
        return EnrichedDiffConflict(
            uuid=str(diff_conflict_node.get("uuid")),
            base_branch_action=DiffAction(str(diff_conflict_node.get("base_branch_action"))),
            base_branch_value=base_branch_value,
            base_branch_changed_at=Timestamp(base_timestamp_str) if base_timestamp_str else None,
            base_branch_label=base_branch_label,
            diff_branch_action=DiffAction(str(diff_conflict_node.get("diff_branch_action"))),
            diff_branch_value=diff_branch_value,
            diff_branch_label=diff_branch_label,
            diff_branch_changed_at=Timestamp(diff_timestamp_str) if diff_timestamp_str else None,
            selected_branch=ConflictSelection(selected_branch) if selected_branch else None,
            resolvable=resolvable,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.constants import RelationshipHierarchyDirection
from infrahub.core.manager import NodeManager
from infrahub.core.schema import NodeSchema
from infrahub.events.models import EventNode

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase


async def collect_ancestors(db: InfrahubDatabase, branch: Branch, node_kind: str, node_id: str) -> list[EventNode]:
    schema = db.schema.get(name=node_kind, branch=branch, duplicate=False)

    if not isinstance(schema, NodeSchema):
        return []

    ancestors = await NodeManager.query_hierarchy(
        db=db,
        branch=branch,
        direction=RelationshipHierarchyDirection.ANCESTORS,
        id=node_id,
        node_schema=schema,
        filters={"id": None},
    )
    return [EventNode(id=ancestor.get_id(), kind=ancestor.get_kind()) for ancestor in ancestors.values()]



from graphene import Field, InputObjectType, Int, List, NonNull, ObjectType, String
from graphene.types.uuid import UUID

from .enums import Severity


class RelatedTaskLogCreateInput(InputObjectType):
    message = String(required=True)
    severity = Severity(required=True)


class TaskLogCreateInput(RelatedTaskLogCreateInput):
    task_id = UUID(required=True)


class TaskLog(ObjectType):
    message = String(required=True)
    severity = String(required=True)
    task_id = String(required=True)
    timestamp = String(required=True)
    id = String(required=False)


class TaskLogNodes(ObjectType):
    node = Field(TaskLog)


class TaskLogEdge(ObjectType):
    edges = List(NonNull(TaskLogNodes), required=True)
    count = Int(required=True)



from infrahub import config
from infrahub.auth import AccountSession
from infrahub.core.account import GlobalPermission
from infrahub.core.branch import Branch
from infrahub.core.constants import GlobalPermissions, PermissionDecision
from infrahub.database import InfrahubDatabase
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.initialization import GraphqlParams

from .interface import CheckerResolution, GraphQLQueryPermissionCheckerInterface


class SuperAdminPermissionChecker(GraphQLQueryPermissionCheckerInterface):
    """Checker allows a user to do anything (if the checker runs first)."""

    permission_required = GlobalPermission(
        action=GlobalPermissions.SUPER_ADMIN.value, decision=PermissionDecision.ALLOW_ALL.value
    )

    async def supports(self, db: InfrahubDatabase, account_session: AccountSession, branch: Branch) -> bool:  # noqa: ARG002
        return config.SETTINGS.main.allow_anonymous_access or account_session.authenticated

    async def check(
        self,
        db: InfrahubDatabase,  # noqa: ARG002
        account_session: AccountSession,  # noqa: ARG002
        analyzed_query: InfrahubGraphQLQueryAnalyzer,  # noqa: ARG002
        query_parameters: GraphqlParams,
        branch: Branch,  # noqa: ARG002
    ) -> CheckerResolution:
        return (
            CheckerResolution.TERMINATE
            if query_parameters.context.active_permissions.has_permission(permission=self.permission_required)
            else CheckerResolution.NEXT_CHECKER
        )



from infrahub.utils import InfrahubStringEnum


class WorkflowType(InfrahubStringEnum):
    INTERNAL = "internal"
    CORE = "core"
    USER = "user"


TAG_NAMESPACE = "infrahub.app"


class WorkflowTag(InfrahubStringEnum):
    BRANCH = "branch/{identifier}"
    WORKFLOWTYPE = "workflow-type/{identifier}"
    DATABASE_CHANGE = "database-change"
    RELATED_NODE = "node/{identifier}"

    def render(self, identifier: str | None = None) -> str:
        if identifier is None:
            return f"{TAG_NAMESPACE}/{self.value}"
        rendered_value = str(self.value).format(identifier=identifier)
        return f"{TAG_NAMESPACE}/{rendered_value}"



import importlib.metadata

from .constants import InfrahubType


def determine_infrahub_type() -> InfrahubType:
    try:
        importlib.metadata.version("infrahub-enterprise")
        return InfrahubType.ENTERPRISE
    except importlib.metadata.PackageNotFoundError:
        return InfrahubType.COMMUNITY



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphql.type.definition import GraphQLNonNull
from infrahub_sdk.utils import extract_fields
from opentelemetry import trace

from infrahub.core.constants import BranchSupportType, InfrahubKind, RelationshipHierarchyDirection
from infrahub.core.manager import NodeManager
from infrahub.exceptions import NodeNotFoundError

from ..models import OrderModel
from ..parser import extract_selection
from ..permissions import get_permissions

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.schema import NodeSchema
    from infrahub.graphql.initialization import GraphqlContext


@trace.get_tracer(__name__).start_as_current_span("account_resolver")
async def account_resolver(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
) -> dict:
    fields = await extract_fields(info.field_nodes[0].selection_set)
    graphql_context: GraphqlContext = info.context

    async with graphql_context.db.start_session(read_only=True) as db:
        results = await NodeManager.query(
            schema=InfrahubKind.GENERICACCOUNT,
            filters={"ids": [graphql_context.account_session.account_id]},
            fields=fields,
            db=db,
            order=OrderModel(disable=True),
        )
        if results:
            account_profile = await results[0].to_graphql(db=db, fields=fields)
            return account_profile

        raise NodeNotFoundError(
            node_type=InfrahubKind.GENERICACCOUNT, identifier=graphql_context.account_session.account_id
        )


@trace.get_tracer(__name__).start_as_current_span("default_resolver")
async def default_resolver(*args: Any, **kwargs) -> dict | list[dict] | None:
    """Not sure why but the default resolver returns sometime 4 positional args and sometime 2.

    When it returns 4, they are organized as follow
        - field name
        - ???
        - parent
        - info
    When it returns 2, they are organized as follow
        - parent
        - info
    """

    parent = None
    info = None
    field_name = None

    if len(args) == 4:
        parent = args[2]
        info = args[3]
        field_name = args[0]
    elif len(args) == 2:
        parent = args[0]
        info = args[1]
        field_name = info.field_name
    else:
        raise ValueError(f"expected either 2 or 4 args for default_resolver, got {len(args)}")

    # Extract the InfraHub schema by inspecting the GQL Schema
    node_schema: NodeSchema = (
        info.parent_type.of_type.graphene_type._meta.schema
        if isinstance(info.parent_type, GraphQLNonNull)
        else info.parent_type.graphene_type._meta.schema
    )

    # If the field is an attribute, return its value directly
    if field_name not in node_schema.relationship_names:
        return parent.get(field_name, None)

    # Extract the contextual information from the request context
    graphql_context: GraphqlContext = info.context

    # Extract the name of the fields in the GQL query
    fields = await extract_fields(info.field_nodes[0].selection_set)

    # Extract the schema of the node on the other end of the relationship from the GQL Schema
    node_rel = node_schema.get_relationship(info.field_name)

    # Extract only the filters from the kwargs and prepend the name of the field to the filters
    filters = {
        f"{info.field_name}__{key}": value
        for key, value in kwargs.items()
        if ("__" in key and value) or key in ["id", "ids"]
    }

    async with graphql_context.db.start_session(read_only=True) as db:
        objs = await NodeManager.query_peers(
            db=db,
            ids=[parent["id"]],
            source_kind=node_schema.kind,
            schema=node_rel,
            filters=filters,
            fields=fields,
            at=graphql_context.at,
            branch=graphql_context.branch,
            branch_agnostic=node_rel.branch is BranchSupportType.AGNOSTIC,
            fetch_peers=True,
        )

        if node_rel.cardinality == "many":
            return [
                await obj.to_graphql(db=db, fields=fields, related_node_ids=graphql_context.related_node_ids)
                for obj in objs
            ]

        # If cardinality is one
        if not objs:
            return None

        return await objs[0].to_graphql(db=db, fields=fields, related_node_ids=graphql_context.related_node_ids)


@trace.get_tracer(__name__).start_as_current_span("parent_field_name_resolver")
async def parent_field_name_resolver(parent: dict[str, dict], info: GraphQLResolveInfo) -> dict:
    """This resolver gets used when we know that the parent resolver has already gathered the required information.

    An example of this is the permissions field at the top level within default_paginated_list_resolver()
    """

    return parent[info.field_name]


@trace.get_tracer(__name__).start_as_current_span("default_paginated_list_resolver")
async def default_paginated_list_resolver(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
    offset: int | None = None,
    limit: int | None = None,
    order: OrderModel | None = None,
    partial_match: bool = False,
    **kwargs: dict[str, Any],
) -> dict[str, Any]:
    schema: NodeSchema = (
        info.return_type.of_type.graphene_type._meta.schema
        if isinstance(info.return_type, GraphQLNonNull)
        else info.return_type.graphene_type._meta.schema
    )

    fields = await extract_selection(info.field_nodes[0], schema=schema)

    graphql_context: GraphqlContext = info.context
    async with graphql_context.db.start_session(read_only=True) as db:
        response: dict[str, Any] = {"edges": []}
        filters = {
            key: value for key, value in kwargs.items() if ("__" in key and value is not None) or key in ("ids", "hfid")
        }

        edges = fields.get("edges", {})
        node_fields = edges.get("node", {})

        permission_set: dict[str, Any] | None = None
        permissions = (
            await get_permissions(schema=schema, graphql_context=graphql_context)
            if graphql_context.permissions
            else None
        )
        if fields.get("permissions"):
            response["permissions"] = permissions

        if permissions:
            for edge in permissions["edges"]:
                if edge["node"]["kind"] == schema.kind:
                    permission_set = edge["node"]

        objs = []
        if edges or "hfid" in filters:
            objs = await NodeManager.query(
                db=db,
                schema=schema,
                filters=filters or None,
                fields=node_fields,
                at=graphql_context.at,
                branch=graphql_context.branch,
                limit=limit,
                offset=offset,
                account=graphql_context.account_session,
                include_source=True,
                include_owner=True,
                partial_match=partial_match,
                order=order,
            )

        if "count" in fields:
            if filters.get("hfid"):
                response["count"] = len(objs)
            else:
                response["count"] = await NodeManager.count(
                    db=db,
                    schema=schema,
                    filters=filters,
                    at=graphql_context.at,
                    branch=graphql_context.branch,
                    partial_match=partial_match,
                )

        # get peer IDs for relationships
        if objs:
            objects = [
                {
                    "node": await obj.to_graphql(
                        db=db,
                        fields=node_fields,
                        related_node_ids=graphql_context.related_node_ids,
                        permissions=permission_set,
                    )
                }
                for obj in objs
            ]
            response["edges"] = objects

        return response


@trace.get_tracer(__name__).start_as_current_span("single_relationship_resolver")
async def single_relationship_resolver(parent: dict, info: GraphQLResolveInfo, **kwargs: Any) -> dict[str, Any]:
    graphql_context: GraphqlContext = info.context
    resolver = graphql_context.single_relationship_resolver
    return await resolver.resolve(parent=parent, info=info, **kwargs)


@trace.get_tracer(__name__).start_as_current_span("many_relationship_resolver")
async def many_relationship_resolver(
    parent: dict, info: GraphQLResolveInfo, include_descendants: bool | None = False, **kwargs: Any
) -> dict[str, Any]:
    graphql_context: GraphqlContext = info.context
    resolver = graphql_context.many_relationship_resolver
    return await resolver.resolve(parent=parent, info=info, include_descendants=include_descendants, **kwargs)


async def ancestors_resolver(parent: dict, info: GraphQLResolveInfo, **kwargs) -> dict[str, Any]:
    return await hierarchy_resolver(
        direction=RelationshipHierarchyDirection.ANCESTORS, parent=parent, info=info, **kwargs
    )


async def descendants_resolver(parent: dict, info: GraphQLResolveInfo, **kwargs) -> dict[str, Any]:
    return await hierarchy_resolver(
        direction=RelationshipHierarchyDirection.DESCENDANTS, parent=parent, info=info, **kwargs
    )


@trace.get_tracer(__name__).start_as_current_span("hierarchy_resolver")
async def hierarchy_resolver(
    direction: RelationshipHierarchyDirection, parent: dict, info: GraphQLResolveInfo, **kwargs
) -> dict[str, Any]:
    """Resolver for ancestors and dependants for Hierarchical nodes

    This resolver is used for paginated responses and as such we redefined the requested
    fields by only reusing information below the 'node' key.
    """
    # Extract the InfraHub schema by inspecting the GQL Schema
    node_schema: NodeSchema = (
        info.parent_type.of_type.graphene_type._meta.schema
        if isinstance(info.parent_type, GraphQLNonNull)
        else info.parent_type.graphene_type._meta.schema
    )

    graphql_context: GraphqlContext = info.context

    # Extract the name of the fields in the GQL query
    fields = await extract_fields(info.field_nodes[0].selection_set)
    edges = fields.get("edges", {})
    node_fields = edges.get("node", {})

    # Extract only the filters from the kwargs and prepend the name of the field to the filters
    offset = kwargs.pop("offset", None)
    limit = kwargs.pop("limit", None)

    filters = {
        f"{info.field_name}__{key}": value
        for key, value in kwargs.items()
        if ("__" in key and value) or key in ["id", "ids"]
    }

    response: dict[str, Any] = {"edges": [], "count": None}

    async with graphql_context.db.start_session(read_only=True) as db:
        if "count" in fields:
            response["count"] = await NodeManager.count_hierarchy(
                db=db,
                id=parent["id"],
                direction=direction,
                node_schema=node_schema,
                filters=filters,
                at=graphql_context.at,
                branch=graphql_context.branch,
            )

        if not node_fields:
            return response

        objs = await NodeManager.query_hierarchy(
            db=db,
            id=parent["id"],
            direction=direction,
            node_schema=node_schema,
            filters=filters,
            fields=node_fields,
            offset=offset,
            limit=limit,
            at=graphql_context.at,
            branch=graphql_context.branch,
        )

        if not objs:
            return response
        node_graph = [await obj.to_graphql(db=db, fields=node_fields) for obj in objs.values()]

        entries = []
        for node in node_graph:
            entry = {"node": {}, "properties": {}}
            entry["node"] = node
            entries.append(entry)
        response["edges"] = entries

        return response



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ..exceptions import (
    Error,
)
from ..protocols_base import CoreNodeBase
from .constants import PROPERTIES_FLAG, PROPERTIES_OBJECT

if TYPE_CHECKING:
    from ..client import InfrahubClient, InfrahubClientSync
    from ..schema import RelationshipSchemaAPI
    from .node import InfrahubNode, InfrahubNodeSync


class RelatedNodeBase:
    """Base class for representing a related node in a relationship."""

    def __init__(self, branch: str, schema: RelationshipSchemaAPI, data: Any | dict, name: str | None = None):
        """
        Args:
            branch (str): The branch where the related node resides.
            schema (RelationshipSchema): The schema of the relationship.
            data (Union[Any, dict]): Data representing the related node.
            name (Optional[str]): The name of the related node.
        """
        self.schema = schema
        self.name = name

        self._branch = branch

        self._properties_flag = PROPERTIES_FLAG
        self._properties_object = PROPERTIES_OBJECT
        self._properties = self._properties_flag + self._properties_object

        self._peer = None
        self._id: str | None = None
        self._hfid: list[str] | None = None
        self._display_label: str | None = None
        self._typename: str | None = None
        self._kind: str | None = None

        if isinstance(data, (CoreNodeBase)):
            self._peer = data
            for prop in self._properties:
                setattr(self, prop, None)

        elif isinstance(data, list):
            data = {"hfid": data}
        elif not isinstance(data, dict):
            data = {"id": data}

        if isinstance(data, dict):
            # To support both with and without pagination, we split data into node_data and properties_data
            # We should probably clean that once we'll remove the code without pagination.
            node_data = data.get("node", data)
            properties_data = data.get("properties", data)

            if node_data:
                self._id = node_data.get("id", None)
                self._hfid = node_data.get("hfid", None)
                self._kind = node_data.get("kind", None)
                self._display_label = node_data.get("display_label", None)
                self._typename = node_data.get("__typename", None)

            self.updated_at: str | None = data.get("updated_at", data.get("_relation__updated_at", None))

            # FIXME, we won't need that once we are only supporting paginated results
            if self._typename and self._typename.startswith("Related"):
                self._typename = self._typename[7:]

            for prop in self._properties:
                prop_data = properties_data.get(prop, properties_data.get(f"_relation__{prop}", None))
                if prop_data and isinstance(prop_data, dict) and "id" in prop_data:
                    setattr(self, prop, prop_data["id"])
                elif prop_data and isinstance(prop_data, (str, bool)):
                    setattr(self, prop, prop_data)
                else:
                    setattr(self, prop, None)

    @property
    def id(self) -> str | None:
        if self._peer:
            return self._peer.id
        return self._id

    @property
    def hfid(self) -> list[Any] | None:
        if self._peer:
            return self._peer.hfid
        return self._hfid

    @property
    def hfid_str(self) -> str | None:
        if self._peer and self.hfid:
            return self._peer.get_human_friendly_id_as_string(include_kind=True)
        return None

    @property
    def is_resource_pool(self) -> bool:
        if self._peer:
            return self._peer.is_resource_pool()
        return False

    @property
    def initialized(self) -> bool:
        return bool(self.id) or bool(self.hfid)

    @property
    def display_label(self) -> str | None:
        if self._peer:
            return self._peer.display_label
        return self._display_label

    @property
    def typename(self) -> str | None:
        if self._peer:
            return self._peer.typename
        return self._typename

    @property
    def kind(self) -> str | None:
        if self._peer:
            return self._peer.get_kind()
        return self._kind

    def _generate_input_data(self, allocate_from_pool: bool = False) -> dict[str, Any]:
        data: dict[str, Any] = {}

        if self.is_resource_pool and allocate_from_pool:
            return {"from_pool": {"id": self.id}}

        if self.id is not None:
            data["id"] = self.id
        elif self.hfid is not None:
            data["hfid"] = self.hfid
            if self._kind is not None:
                data["kind"] = self._kind

        for prop_name in self._properties:
            if getattr(self, prop_name) is not None:
                data[f"_relation__{prop_name}"] = getattr(self, prop_name)

        return data

    def _generate_mutation_query(self) -> dict[str, Any]:
        if self.name and self.is_resource_pool:
            # If a related node points to a pool, ask for the ID of the pool allocated resource
            return {self.name: {"node": {"id": None, "display_label": None, "__typename": None}}}
        return {}

    @classmethod
    def _generate_query_data(cls, peer_data: dict[str, Any] | None = None, property: bool = False) -> dict:
        """Generates the basic structure of a GraphQL query for a single relationship.

        Args:
            peer_data (dict[str, Union[Any, Dict]], optional): Additional data to be included in the query for the node.
                This is used to add extra fields when prefetching related node data.

        Returns:
            Dict: A dictionary representing the basic structure of a GraphQL query, including the node's ID, display label,
                and typename. The method also includes additional properties and any peer_data provided.
        """
        data: dict[str, Any] = {"node": {"id": None, "hfid": None, "display_label": None, "__typename": None}}
        properties: dict[str, Any] = {}

        if property:
            for prop_name in PROPERTIES_FLAG:
                properties[prop_name] = None
            for prop_name in PROPERTIES_OBJECT:
                properties[prop_name] = {"id": None, "display_label": None, "__typename": None}

        if properties:
            data["properties"] = properties
        if peer_data:
            data["node"].update(peer_data)

        return data


class RelatedNode(RelatedNodeBase):
    """Represents a RelatedNodeBase in an asynchronous context."""

    def __init__(
        self,
        client: InfrahubClient,
        branch: str,
        schema: RelationshipSchemaAPI,
        data: Any | dict,
        name: str | None = None,
    ):
        """
        Args:
            client (InfrahubClient): The client used to interact with the backend asynchronously.
            branch (str): The branch where the related node resides.
            schema (RelationshipSchema): The schema of the relationship.
            data (Union[Any, dict]): Data representing the related node.
            name (Optional[str]): The name of the related node.
        """
        self._client = client
        super().__init__(branch=branch, schema=schema, data=data, name=name)

    async def fetch(self, timeout: int | None = None) -> None:
        if not self.id or not self.typename:
            raise Error("Unable to fetch the peer, id and/or typename are not defined")

        self._peer = await self._client.get(
            kind=self.typename, id=self.id, populate_store=True, branch=self._branch, timeout=timeout
        )

    @property
    def peer(self) -> InfrahubNode:
        return self.get()

    def get(self) -> InfrahubNode:
        if self._peer:
            return self._peer  # type: ignore[return-value]

        if self.id and self.typename:
            return self._client.store.get(key=self.id, kind=self.typename, branch=self._branch)  # type: ignore[return-value]

        if self.hfid_str:
            return self._client.store.get(key=self.hfid_str, branch=self._branch)  # type: ignore[return-value]

        raise ValueError("Node must have at least one identifier (ID or HFID) to query it.")


class RelatedNodeSync(RelatedNodeBase):
    """Represents a related node in a synchronous context."""

    def __init__(
        self,
        client: InfrahubClientSync,
        branch: str,
        schema: RelationshipSchemaAPI,
        data: Any | dict,
        name: str | None = None,
    ):
        """
        Args:
            client (InfrahubClientSync): The client used to interact with the backend synchronously.
            branch (str): The branch where the related node resides.
            schema (RelationshipSchema): The schema of the relationship.
            data (Union[Any, dict]): Data representing the related node.
            name (Optional[str]): The name of the related node.
        """
        self._client = client
        super().__init__(branch=branch, schema=schema, data=data, name=name)

    def fetch(self, timeout: int | None = None) -> None:
        if not self.id or not self.typename:
            raise Error("Unable to fetch the peer, id and/or typename are not defined")

        self._peer = self._client.get(
            kind=self.typename, id=self.id, populate_store=True, branch=self._branch, timeout=timeout
        )

    @property
    def peer(self) -> InfrahubNodeSync:
        return self.get()

    def get(self) -> InfrahubNodeSync:
        if self._peer:
            return self._peer  # type: ignore[return-value]

        if self.id and self.typename:
            return self._client.store.get(key=self.id, kind=self.typename, branch=self._branch)  # type: ignore[return-value]

        if self.hfid_str:
            return self._client.store.get(key=self.hfid_str, branch=self._branch)  # type: ignore[return-value]

        raise ValueError("Node must have at least one identifier (ID or HFID) to query it.")



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncIterator

from fastapi import Depends, Query, Request
from fastapi.security import APIKeyHeader, HTTPAuthorizationCredentials, HTTPBearer
from pydantic import BaseModel, ConfigDict

from infrahub import config
from infrahub.auth import AccountSession, authentication_token, validate_jwt_access_token, validate_jwt_refresh_token
from infrahub.context import InfrahubContext
from infrahub.core.branch import Branch  # noqa: TC001
from infrahub.core.registry import registry
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase  # noqa: TC001
from infrahub.exceptions import AuthorizationError
from infrahub.permissions import PermissionManager

if TYPE_CHECKING:
    from neo4j import AsyncSession

    from infrahub.models import RefreshTokenData

jwt_scheme = HTTPBearer(auto_error=False)
api_key_scheme = APIKeyHeader(name="X-INFRAHUB-KEY", auto_error=False)


async def cookie_auth_scheme(request: Request) -> str | None:
    return request.cookies.get("access_token")  # Replace with the actual name of your JWT cookie


class BranchParams(BaseModel):
    branch: Branch
    at: Timestamp

    model_config = ConfigDict(arbitrary_types_allowed=True)


async def get_session(request: Request) -> AsyncIterator[AsyncSession]:
    session = request.app.state.db.session(database=config.SETTINGS.database.database_name)
    try:
        yield session
    finally:
        await session.close()


async def get_db(request: Request) -> InfrahubDatabase:
    return request.app.state.db.start_session()


async def get_access_token(
    request: Request, jwt_header: HTTPAuthorizationCredentials = Depends(jwt_scheme)
) -> AccountSession:
    if jwt_header:
        return await validate_jwt_access_token(token=jwt_header.credentials)
    if token := request.cookies.get("access_token"):
        return await validate_jwt_access_token(token=token)

    raise AuthorizationError("A JWT access token is required to perform this operation.")


async def get_refresh_token(
    request: Request,
    db: InfrahubDatabase = Depends(get_db),
    jwt_header: HTTPAuthorizationCredentials | None = Depends(jwt_scheme),
) -> RefreshTokenData:
    token = None

    # Check for token in header
    if jwt_header:
        token = jwt_header.credentials

    # If no auth header, try to get the token from the cookie
    if not token:
        token = request.cookies.get("refresh_token")

    # If still no token, raise an error
    if not token:
        raise AuthorizationError("A JWT refresh token is required to perform this operation.")

    return await validate_jwt_refresh_token(db=db, token=token)


async def get_branch_params(
    db: InfrahubDatabase = Depends(get_db),
    branch_name: str | None = Query(None, alias="branch", description="Name of the branch to use for the query"),
    at: str | None = Query(None, description="Time to use for the query, in absolute or relative format"),
) -> BranchParams:
    branch = await registry.get_branch(db=db, branch=branch_name)

    return BranchParams(branch=branch, at=Timestamp(at))


async def get_branch_dep(
    db: InfrahubDatabase = Depends(get_db),
    branch_name: str | None = Query(None, alias="branch", description="Name of the branch to use for the query"),
) -> Branch:
    return await registry.get_branch(db=db, branch=branch_name)


async def get_current_user(
    request: Request,
    jwt_header: HTTPAuthorizationCredentials = Depends(jwt_scheme),
    db: InfrahubDatabase = Depends(get_db),
    api_key: str = Depends(api_key_scheme),
) -> AccountSession:
    """Return current user"""
    jwt_token = None
    if jwt_header:
        jwt_token = jwt_header.credentials

    if not jwt_token:
        jwt_token = request.cookies.get("access_token")

    account_session = await authentication_token(db=db, jwt_token=jwt_token, api_key=api_key)

    if (
        account_session.authenticated
        or request.url.path.startswith("/graphql")
        or (config.SETTINGS.main.allow_anonymous_access and request.method.lower() in ["get", "options"])
    ):
        return account_session

    raise AuthorizationError("Authentication is required")


async def get_permission_manager(
    db: InfrahubDatabase = Depends(get_db),
    branch_params: BranchParams = Depends(get_branch_params),
    account_session: AccountSession = Depends(get_current_user),
) -> PermissionManager:
    """Return a `PermissionManager` for an account session based on a branch."""
    permission_manager = PermissionManager(account_session=account_session)
    await permission_manager.load_permissions(db=db, branch=branch_params.branch)

    return permission_manager


async def get_context(
    branch: Branch = Depends(get_branch_dep),
    account_session: AccountSession = Depends(get_current_user),
) -> InfrahubContext:
    return InfrahubContext.init(branch=branch, account=account_session)



from typing import Iterable

from infrahub.core.constants import InfrahubKind
from infrahub.database import InfrahubDatabase


class IpamKindsGetter:
    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db

    async def get_ipam_address_kinds(self, branch_names: Iterable[str]) -> set[str]:
        ip_address_kinds: set[str] = set()
        for branch_name in branch_names:
            address_generic_schema_source = self.db.schema.get(
                InfrahubKind.IPADDRESS, branch=branch_name, duplicate=False
            )
            address_generic_schema_target = self.db.schema.get(
                InfrahubKind.IPADDRESS, branch=branch_name, duplicate=False
            )

            ip_address_kinds.update(
                set(
                    getattr(address_generic_schema_target, "used_by", [])
                    + getattr(address_generic_schema_source, "used_by", [])
                )
            )
        return ip_address_kinds

    async def get_ipam_prefix_kinds(self, branch_names: Iterable[str]) -> set[str]:
        ip_prefix_kinds: set[str] = set()
        for branch_name in branch_names:
            prefix_generic_schema_source = self.db.schema.get(
                InfrahubKind.IPPREFIX, branch=branch_name, duplicate=False
            )
            prefix_generic_schema_target = self.db.schema.get(
                InfrahubKind.IPPREFIX, branch=branch_name, duplicate=False
            )

            ip_prefix_kinds.update(
                set(
                    getattr(prefix_generic_schema_source, "used_by", [])
                    + getattr(prefix_generic_schema_target, "used_by", [])
                )
            )
        return ip_prefix_kinds



from prometheus_client import Histogram

METRIC_PREFIX = "infrahub_graphql"

SCHEMA_GENERATE_GRAPHQL_METRICS = Histogram(
    f"{METRIC_PREFIX}_generate_schema",
    "Time to generate the GraphQL Schema",
    labelnames=["branch"],
    buckets=[0.0005, 0.25, 0.5, 1, 5],
)

GRAPHQL_DURATION_METRICS = Histogram(
    f"{METRIC_PREFIX}_duration_seconds",
    "GraphQL query duration, in seconds",
    labelnames=["type", "operation", "branch", "name", "query_id"],
    buckets=[0.01, 0.025, 0.05, 0.1, 0.5, 1, 5, 10, 60],
)
GRAPHQL_RESPONSE_SIZE_METRICS = Histogram(
    f"{METRIC_PREFIX}_response_size_bytes",
    "GraphQL query response size, in bytes",
    labelnames=["type", "operation", "branch", "name", "query_id"],
    buckets=[500, 1000, 5000, 10000, 100000, 1000000],
)
GRAPHQL_QUERY_DEPTH_METRICS = Histogram(
    f"{METRIC_PREFIX}_query_depth",
    "GraphQL query depth",
    labelnames=["type", "operation", "branch", "name", "query_id"],
    buckets=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 25, 50, 100],
)
GRAPHQL_QUERY_HEIGHT_METRICS = Histogram(
    f"{METRIC_PREFIX}_query_height",
    "GraphQL query height",
    labelnames=["type", "operation", "branch", "name", "query_id"],
    buckets=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 25, 50, 100],
)
GRAPHQL_QUERY_VARS_METRICS = Histogram(
    f"{METRIC_PREFIX}_variables",
    "GraphQL query number of variables",
    labelnames=["type", "operation", "branch", "name", "query_id"],
    buckets=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 25, 50, 100],
)
GRAPHQL_QUERY_ERRORS_METRICS = Histogram(
    f"{METRIC_PREFIX}_errors",
    "GraphQL query number of errors",
    labelnames=["type", "operation", "branch", "name", "query_id"],
    buckets=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 25, 50, 100],
)
GRAPHQL_QUERY_OBJECTS_METRICS = Histogram(
    f"{METRIC_PREFIX}_query_objects",
    "GraphQL number of objects in the query",
    labelnames=["type", "operation", "branch", "name", "query_id"],
    buckets=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 25, 50, 100],
)
GRAPHQL_TOP_LEVEL_QUERIES_METRICS = Histogram(
    f"{METRIC_PREFIX}_top_level_queries",
    "GraphQL number of top level queries",
    labelnames=["type", "operation", "branch", "name", "query_id"],
    buckets=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 25, 50, 100],
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import BigInt, Field, Float, Int, List, NonNull, ObjectType, String
from infrahub_sdk.utils import extract_fields_first_node

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.ipam.utilization import PrefixUtilizationGetter
from infrahub.core.manager import NodeManager
from infrahub.core.query.ipam import IPPrefixUtilization
from infrahub.core.query.resource_manager import (
    IPAddressPoolGetIdentifiers,
    NumberPoolGetAllocated,
    PrefixPoolGetIdentifiers,
)
from infrahub.exceptions import NodeNotFoundError, SchemaNotFoundError, ValidationError
from infrahub.pools.number import NumberUtilizationGetter

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.protocols import CoreNode
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext


class IPPoolUtilizationResource(ObjectType):
    id = Field(String, required=True, description="The ID of the current resource")
    display_label = Field(String, required=True, description="The common name of the resource")
    kind = Field(String, required=True, description="The resource kind")
    weight = Field(BigInt, required=True, description="The relative weight of this resource.")
    utilization = Field(Float, required=True, description="The overall utilization of the resource.")
    utilization_branches = Field(
        Float, required=True, description="The utilization of the resource on all non default branches."
    )
    utilization_default_branch = Field(
        Float, required=True, description="The overall utilization of the resource isolated to the default branch."
    )


class IPPrefixUtilizationEdge(ObjectType):
    node = Field(IPPoolUtilizationResource, required=True)


class PoolAllocatedNode(ObjectType):
    id = Field(String, required=True, description="The ID of the allocated node")
    display_label = Field(String, required=True, description="The common name of the resource")
    kind = Field(String, required=True, description="The node kind")
    branch = Field(String, required=True, description="The branch where the node is allocated")
    identifier = Field(String, required=False, description="Identifier used for the allocation")


class PoolAllocatedEdge(ObjectType):
    node = Field(PoolAllocatedNode, required=True)


def _validate_pool_type(pool_id: str, pool: CoreNode | None = None) -> CoreNode:
    if not pool or pool.get_kind() not in [
        InfrahubKind.IPADDRESSPOOL,
        InfrahubKind.IPPREFIXPOOL,
        InfrahubKind.NUMBERPOOL,
    ]:
        raise NodeNotFoundError(node_type="ResourcePool", identifier=pool_id)
    return pool


class PoolAllocated(ObjectType):
    count = Field(BigInt, required=True, description="The number of allocations within the selected pool.")
    edges = Field(List(of_type=NonNull(PoolAllocatedEdge), required=True), required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        pool_id: str,
        resource_id: str,
        offset: int = 0,
        limit: int = 10,
    ) -> dict:
        graphql_context: GraphqlContext = info.context
        pool: CoreNode | None = await NodeManager.get_one(
            id=pool_id, db=graphql_context.db, branch=graphql_context.branch
        )

        fields = await extract_fields_first_node(info=info)

        allocated_kinds: list[str] = []
        pool = _validate_pool_type(pool_id=pool_id, pool=pool)
        match pool.get_kind():
            case InfrahubKind.NUMBERPOOL:
                return await resolve_number_pool_allocation(
                    db=graphql_context.db,
                    graphql_context=graphql_context,
                    pool=pool,
                    fields=fields,
                    offset=offset,
                    limit=limit,
                )
            case InfrahubKind.IPPREFIXPOOL:
                allocated_kinds.append(InfrahubKind.IPPREFIX)
            case InfrahubKind.IPADDRESSPOOL:
                allocated_kinds.append(InfrahubKind.IPADDRESS)

        resources = await pool.resources.get_peers(db=graphql_context.db)  # type: ignore[attr-defined,union-attr]
        if resource_id not in resources:
            raise ValidationError(
                input_value=f"The selected pool_id={pool_id} doesn't contain the requested resource_id={resource_id}"
            )

        resource = resources[resource_id]

        query = await IPPrefixUtilization.init(
            db=graphql_context.db,
            at=graphql_context.at,
            ip_prefixes=[resource],
            allocated_kinds=allocated_kinds,
            offset=offset,
            limit=limit,
        )
        response: dict[str, Any] = {}
        if "count" in fields:
            response["count"] = await query.count(db=graphql_context.db)

        if edges := fields.get("edges"):
            await query.execute(db=graphql_context.db)

            node_fields = edges.get("node", {})

            nodes = []
            for result in query.get_results():
                child_node = result.get_node("child")
                child_value_node = result.get_node("av")
                node_id = str(child_node.get("uuid"))

                child_ip_value = child_value_node.get("value")
                kind = child_node.get("kind")
                branch_name = str(result.get("branch"))

                nodes.append(
                    {"node": {"id": node_id, "kind": kind, "branch": branch_name, "display_label": child_ip_value}}
                )

            if "identifier" in node_fields:
                allocated_ids = [node["node"]["id"] for node in nodes]
                identifier_query_map = {
                    InfrahubKind.IPADDRESSPOOL: IPAddressPoolGetIdentifiers,
                    InfrahubKind.IPPREFIXPOOL: PrefixPoolGetIdentifiers,
                }
                identifier_query_class = identifier_query_map.get(pool.get_kind())
                if not identifier_query_class:
                    raise ValidationError(input_value=f"This query doesn't get support {pool.get_kind()}")
                identifier_query = await identifier_query_class.init(
                    db=graphql_context.db, at=graphql_context.at, pool_id=pool_id, allocated=allocated_ids
                )
                await identifier_query.execute(db=graphql_context.db)

                reservations = {}
                for result in identifier_query.get_results():
                    reservation = result.get_rel("reservation")
                    allocated = result.get_node("allocated")
                    reservations[allocated.get("uuid")] = reservation.get("identifier")

                for node in nodes:
                    node["node"]["identifier"] = reservations.get(node["node"]["id"])

            response["edges"] = nodes

        return response


class PoolUtilization(ObjectType):
    count = Field(BigInt, required=True, description="The number of resources within the selected pool.")
    utilization = Field(Float, required=True, description="The overall utilization of the pool.")
    utilization_branches = Field(Float, required=True, description="The utilization in all non default branches.")
    utilization_default_branch = Field(
        Float, required=True, description="The overall utilization of the pool isolated to the default branch."
    )
    edges = Field(List(of_type=NonNull(IPPrefixUtilizationEdge), required=True), required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        pool_id: str,
    ) -> dict:
        graphql_context: GraphqlContext = info.context
        db: InfrahubDatabase = graphql_context.db
        pool: CoreNode | None = await NodeManager.get_one(id=pool_id, db=db, branch=graphql_context.branch)
        pool = _validate_pool_type(pool_id=pool_id, pool=pool)
        if pool.get_kind() == "CoreNumberPool":
            return await resolve_number_pool_utilization(
                db=db, at=graphql_context.at, pool=pool, branch=graphql_context.branch
            )

        resources_map: dict[str, Node] = {}

        try:
            resources_map = await pool.resources.get_peers(db=db, branch_agnostic=True)  # type: ignore[attr-defined,union-attr]
        except SchemaNotFoundError:
            pass

        utilization_getter = PrefixUtilizationGetter(
            db=db, ip_prefixes=list(resources_map.values()), at=graphql_context.at
        )
        fields = await extract_fields_first_node(info=info)
        response: dict[str, Any] = {}
        total_utilization = None
        default_branch_utilization = None
        if "count" in fields:
            response["count"] = len(resources_map)
        if "utilization" in fields:
            response["utilization"] = total_utilization = await utilization_getter.get_use_percentage()
        if "utilization_default_branch" in fields:
            response["utilization_default_branch"] = (
                default_branch_utilization
            ) = await utilization_getter.get_use_percentage(branch_names=[registry.default_branch])
        if "utilization_branches" in fields:
            total_utilization = (
                total_utilization if total_utilization is not None else await utilization_getter.get_use_percentage()
            )
            default_branch_utilization = (
                default_branch_utilization
                if default_branch_utilization is not None
                else await utilization_getter.get_use_percentage(branch_names=[registry.default_branch])
            )
            response["utilization_branches"] = total_utilization - default_branch_utilization
        if "edges" in fields:
            response["edges"] = []
            if "node" in fields["edges"]:
                node_fields = fields["edges"]["node"]
                for resource_id, resource_node in resources_map.items():
                    resource_total = None
                    default_branch_total = None
                    node_response: dict[str, str | float | int] = {}
                    if "id" in node_fields:
                        node_response["id"] = resource_id
                    if "kind" in node_fields:
                        node_response["kind"] = resource_node.get_kind()
                    if "display_label" in node_fields:
                        node_response["display_label"] = await resource_node.render_display_label(db=db)
                    if "weight" in node_fields:
                        node_response["weight"] = await resource_node.get_resource_weight(db=db)  # type: ignore[attr-defined]
                    if "utilization" in node_fields:
                        node_response["utilization"] = resource_total = await utilization_getter.get_use_percentage(
                            ip_prefixes=[resource_node]
                        )
                    if "utilization_default_branch" in node_fields:
                        node_response["utilization_default_branch"] = (
                            default_branch_total
                        ) = await utilization_getter.get_use_percentage(
                            ip_prefixes=[resource_node], branch_names=[registry.default_branch]
                        )
                    if "utilization_branches" in node_fields:
                        resource_total = (
                            resource_total
                            if resource_total is not None
                            else await utilization_getter.get_use_percentage(ip_prefixes=[resource_node])
                        )
                        default_branch_total = (
                            default_branch_total
                            if default_branch_total is not None
                            else await utilization_getter.get_use_percentage(
                                ip_prefixes=[resource_node], branch_names=[registry.default_branch]
                            )
                        )
                        node_response["utilization_branches"] = resource_total - default_branch_total
                    response["edges"].append({"node": node_response})

        return response


async def resolve_number_pool_allocation(
    db: InfrahubDatabase, graphql_context: GraphqlContext, pool: CoreNode, fields: dict, offset: int, limit: int
) -> dict:
    response: dict[str, Any] = {}
    query = await NumberPoolGetAllocated.init(
        db=db, pool=pool, offset=offset, limit=limit, branch=graphql_context.branch, branch_agnostic=True
    )

    if "count" in fields:
        response["count"] = await query.count(db=db)

    if "edges" in fields:
        await query.execute(db=db)
        edges = []
        for entry in query.results:
            node = {
                "node": {
                    "id": entry.get_as_optional_type("id", str),
                    "kind": pool.node.value,  # type: ignore[attr-defined]
                    "branch": entry.get_as_optional_type("branch", str),
                    "display_label": entry.get_as_optional_type("value", int),
                }
            }
            edges.append(node)

        response["edges"] = edges
    return response


async def resolve_number_pool_utilization(
    db: InfrahubDatabase, pool: CoreNode, at: Timestamp | str | None, branch: Branch
) -> dict:
    """
    Returns a mapping containg utilization info of a number pool.
    The utilization is calculated as the percentage of the total number of values in the pool that are not excluded for the corresponding attribute.
    """

    core_number_pool = await registry.manager.get_one_by_id_or_default_filter(db=db, id=pool.id, kind="CoreNumberPool")
    number_pool = NumberUtilizationGetter(db=db, pool=core_number_pool, at=at, branch=branch)
    await number_pool.load_data()

    return {
        "count": 1,
        "utilization": number_pool.utilization,
        "utilization_default_branch": number_pool.utilization_default_branch,
        "utilization_branches": number_pool.utilization_branches,
        "edges": [
            {
                "node": {
                    "id": pool.get_id(),
                    "kind": "CoreNumberPool",
                    "display_label": pool.name.value,  # type: ignore[attr-defined]
                    "weight": 1,
                    "utilization": number_pool.utilization,
                    "utilization_default_branch": number_pool.utilization_default_branch,
                    "utilization_branches": number_pool.utilization_branches,
                }
            }
        ],
    }


InfrahubResourcePoolAllocated = Field(
    PoolAllocated,
    pool_id=String(required=True),
    resource_id=String(required=True),
    limit=Int(required=False),
    offset=Int(required=False),
    resolver=PoolAllocated.resolve,
    required=True,
)


InfrahubResourcePoolUtilization = Field(
    PoolUtilization, pool_id=String(required=True), resolver=PoolUtilization.resolve, required=True
)



from dataclasses import dataclass, field
from uuid import uuid4

PropertyPrimitives = str | bool | int | float | None


def str_uuid() -> str:
    return str(uuid4())


@dataclass
class VertexToAdd:
    labels: list[str]
    after_props: dict[str, PropertyPrimitives]
    identifier: str = field(default_factory=str_uuid)


@dataclass
class VertexToUpdate:
    db_id: str
    before_props: dict[str, PropertyPrimitives]
    after_props: dict[str, PropertyPrimitives]


@dataclass
class VertexToDelete:
    db_id: str
    labels: list[str]
    before_props: dict[str, PropertyPrimitives]


@dataclass
class EdgeToAdd:
    from_id: str
    to_id: str
    edge_type: str
    after_props: dict[str, PropertyPrimitives]
    identifier: str = field(default_factory=str_uuid)


@dataclass
class EdgeToUpdate:
    db_id: str
    before_props: dict[str, PropertyPrimitives]
    after_props: dict[str, PropertyPrimitives]


@dataclass
class EdgeToDelete:
    db_id: str
    from_id: str
    to_id: str
    edge_type: str
    before_props: dict[str, PropertyPrimitives]


@dataclass
class PatchPlan:
    name: str
    vertices_to_add: list[VertexToAdd] = field(default_factory=list)
    vertices_to_update: list[VertexToUpdate] = field(default_factory=list)
    vertices_to_delete: list[VertexToDelete] = field(default_factory=list)
    edges_to_add: list[EdgeToAdd] = field(default_factory=list)
    edges_to_update: list[EdgeToUpdate] = field(default_factory=list)
    edges_to_delete: list[EdgeToDelete] = field(default_factory=list)
    added_element_db_id_map: dict[str, str] = field(default_factory=dict)
    deleted_db_ids: set[str] = field(default_factory=set)
    reverted_deleted_db_id_map: dict[str, str] = field(default_factory=dict)

    def get_database_id_for_added_element(self, abstract_id: str) -> str:
        return self.added_element_db_id_map.get(abstract_id, abstract_id)

    def has_element_been_added(self, identifier: str) -> bool:
        return identifier in self.added_element_db_id_map

    @property
    def added_vertices(self) -> list[VertexToAdd]:
        return [v for v in self.vertices_to_add if self.has_element_been_added(v.identifier)]

    @property
    def added_edges(self) -> list[EdgeToAdd]:
        return [e for e in self.edges_to_add if self.has_element_been_added(e.identifier)]

    @property
    def deleted_vertices(self) -> list[VertexToDelete]:
        return [v for v in self.vertices_to_delete if v.db_id in self.deleted_db_ids]

    @property
    def deleted_edges(self) -> list[EdgeToDelete]:
        return [e for e in self.edges_to_delete if e.db_id in self.deleted_db_ids]

    def drop_added_db_ids(self, db_ids_to_drop: set[str]) -> None:
        self.added_element_db_id_map = {
            k: v for k, v in self.added_element_db_id_map.items() if v not in db_ids_to_drop
        }

    def drop_deleted_db_ids(self, db_ids_to_drop: set[str]) -> None:
        self.deleted_db_ids -= db_ids_to_drop



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from graphql.language import (
    DirectiveNode,
    FieldNode,
    InlineFragmentNode,
    ListValueNode,
    NameNode,
    SelectionSetNode,
    StringValueNode,
)
from infrahub_sdk.utils import deep_merge_dict

if TYPE_CHECKING:
    from infrahub.core.schema import NodeSchema


@dataclass
class FieldEnricher:
    key: str
    node: FieldNode
    path: str
    fields: dict = field(default_factory=dict)


async def extract_selection(field_node: FieldNode, schema: NodeSchema) -> dict:
    graphql_extractor = GraphQLExtractor(field_node=field_node, schema=schema)
    return await graphql_extractor.get_fields()


class GraphQLExtractor:
    def __init__(self, field_node: FieldNode, schema: NodeSchema) -> None:
        self.field_node = field_node
        self.schema = schema
        self.typename_paths: dict[str, list[FieldEnricher]] = {}
        self.node_path: dict[str, list[FieldEnricher]] = {}

    def _define_node_path(self, path: str) -> None:
        if path not in self.node_path:
            self.node_path[path] = []

    async def get_fields(self) -> dict:
        return await self.extract_fields(selection_set=self.field_node.selection_set) or {}

    def _process_expand_directive(self, path: str, directive: DirectiveNode) -> None:
        excluded_fields = []
        for argument in directive.arguments:
            if argument.name.value == "exclude":
                if isinstance(argument.value, ListValueNode):
                    excluded_fields.extend(
                        [value.value for value in argument.value.values if isinstance(value, StringValueNode)]
                    )

        if path not in self.typename_paths:
            self.typename_paths[path] = []
        self.typename_paths[path].append(
            FieldEnricher(
                key="__typename",
                node=FieldNode(
                    kind="field",
                    name=NameNode(kind="name", value="__typename"),
                    directives=[],
                    arguments=[],
                ),
                path=f"{path}/__typename/",
            )
        )
        if path == "/edges/node/":
            self._define_node_path(path=path)

            self.node_path[path].append(
                FieldEnricher(
                    key="id",
                    node=FieldNode(
                        kind="field",
                        name=NameNode(kind="name", value="id"),
                        directives=[],
                        arguments=[],
                    ),
                    path=f"{path}id/",
                    fields={"id": None},
                )
            )
            attribute_enrichers = []
            attributes = [attribute for attribute in self.schema.attributes if attribute.name not in excluded_fields]
            for attribute in attributes:
                attribute_path = f"{path}{attribute.name}/"
                self._define_node_path(path=attribute_path)
                field_attributes = {"value": None, "is_default": None, "is_from_profile": None}

                enrichers = [
                    FieldEnricher(
                        key=attribute.name,
                        node=FieldNode(
                            kind="field",
                            name=NameNode(
                                kind="name",
                                value=key,
                                directives=[],
                                arguments=[],
                            ),
                        ),
                        path=attribute_path,
                        fields={key: None},
                    )
                    for key in field_attributes
                ]

                self.node_path[attribute_path].extend(enrichers)
                attribute_enrichers.append(
                    FieldNode(
                        kind="field",
                        name=NameNode(kind="name", value=attribute.name),
                        selection_set=SelectionSetNode(selections=tuple(enrichers)),
                    )
                )

            self._define_node_path(path=path)
            self.node_path[path].append(
                FieldEnricher(
                    key="node",
                    path=path,
                    node=FieldNode(
                        kind="field",
                        name=NameNode(kind="name", value="node"),
                        selection_set=SelectionSetNode(selections=tuple(attribute_enrichers)),
                    ),
                    fields={attribute.name: field_attributes for attribute in self.schema.attributes},
                )
            )

    def process_directives(self, node: FieldNode, path: str) -> None:
        for directive in node.directives:
            if directive.name.value == "expand":
                self._process_expand_directive(path=path, directive=directive)

    def apply_directives(self, selection_set: SelectionSetNode, fields: dict, path: str) -> dict:
        if path in self.typename_paths:
            for node in self.typename_paths[path]:
                if "__typename" not in fields:
                    selections = list(selection_set.selections)
                    selections.append(node.node)
                    selection_set.selections = tuple(selections)

        if path in self.node_path:
            for node in self.node_path[path]:
                if node.key not in fields:
                    fields = deep_merge_dict(dicta=fields.copy(), dictb=node.fields)
                    selections = list(selection_set.selections)
                    selections.append(node.node)
                    selection_set.selections = tuple(selections)

            undefined_paths = [key for key in self.node_path if is_child_path(path=path, child=key)]

            for undefined in undefined_paths:
                for sub_node in self.node_path[undefined]:
                    selections = list(selection_set.selections)
                    selections.append(
                        FieldNode(
                            kind="field",
                            name=NameNode(kind="name", value=sub_node.key),
                            selection_set=SelectionSetNode(selections=(sub_node.node,)),
                        )
                    )
                    selection_set.selections = tuple(selections)

            del self.node_path[path]

        return fields

    async def extract_fields(
        self, selection_set: SelectionSetNode | None, path: str = "/"
    ) -> dict[str, dict | None] | None:
        """Extract fields and apply Directives"""
        if not selection_set:
            return None

        fields: dict[str, dict | None] = {}
        for node in selection_set.selections:
            sub_selection_set = getattr(node, "selection_set", None)
            if isinstance(node, FieldNode):
                node_path = f"{path}{node.name.value}/"
                self.process_directives(node=node, path=node_path)

                value = await self.extract_fields(sub_selection_set, path=node_path)
                if node.name.value not in fields:
                    fields[node.name.value] = value
                elif isinstance(fields[node.name.value], dict) and isinstance(value, dict):
                    fields[node.name.value].update(value)  # type: ignore[union-attr]

            elif isinstance(node, InlineFragmentNode):
                for sub_node in node.selection_set.selections:
                    if isinstance(sub_node, FieldNode):
                        sub_node_path = f"{path}{sub_node.name.value}/"
                        sub_sub_selection_set = getattr(sub_node, "selection_set", None)
                        value = await self.extract_fields(sub_sub_selection_set, path=sub_node_path)
                        if sub_node.name.value not in fields:
                            fields[sub_node.name.value] = await self.extract_fields(
                                sub_sub_selection_set, path=sub_node_path
                            )
                        elif isinstance(fields[sub_node.name.value], dict) and isinstance(value, dict):
                            fields[sub_node.name.value].update(value)  # type: ignore[union-attr]

        return self.apply_directives(selection_set=selection_set, fields=fields, path=path)


def is_child_path(path: str, child: str) -> bool:
    if child.startswith(path) and len(child) > len(path):
        return True
    return False



from __future__ import annotations

from typing import TYPE_CHECKING

from graphene import Boolean, Field, InputField, InputObjectType, List, Mutation, NonNull, String

from infrahub.core.manager import NodeManager
from infrahub.generators.models import ProposedChangeGeneratorDefinition, RequestGeneratorDefinitionRun
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.types.context import ContextInput
from infrahub.graphql.types.task import TaskInfo
from infrahub.workflows.catalogue import REQUEST_GENERATOR_DEFINITION_RUN

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from ..initialization import GraphqlContext


class GeneratorDefinitionRequestRunInput(InputObjectType):
    id = InputField(String(required=True), description="ID of the generator definition to run")
    nodes = InputField(List(of_type=NonNull(String)), description="ID list of targets to run the generator for")


class GeneratorDefinitionRequestRun(Mutation):
    class Arguments:
        data = GeneratorDefinitionRequestRunInput(required=True)
        context = ContextInput(required=False)
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: GeneratorDefinitionRequestRunInput,
        context: ContextInput | None = None,
        wait_until_completion: bool = True,
    ) -> GeneratorDefinitionRequestRun:
        graphql_context: GraphqlContext = info.context
        db = graphql_context.db
        await apply_external_context(graphql_context=graphql_context, context_input=context)
        generator_definition = await NodeManager.get_one(
            id=str(data.id), db=db, branch=graphql_context.branch, prefetch_relationships=True, raise_on_error=True
        )
        query = await generator_definition.query.get_peer(db=db)
        repository = await generator_definition.repository.get_peer(db=db)
        group = await generator_definition.targets.get_peer(db=db)

        request_model = RequestGeneratorDefinitionRun(
            generator_definition=ProposedChangeGeneratorDefinition(
                definition_id=generator_definition.id,
                definition_name=generator_definition.name.value,
                class_name=generator_definition.class_name.value,
                file_path=generator_definition.file_path.value,
                query_name=query.name.value,
                query_models=query.models.value,
                repository_id=repository.id,
                parameters=generator_definition.parameters.value,
                group_id=group.id,
                convert_query_response=generator_definition.convert_query_response.value or False,
            ),
            branch=graphql_context.branch.name,
            target_members=data.get("nodes", []),
        )

        if not wait_until_completion:
            workflow = await graphql_context.active_service.workflow.submit_workflow(
                workflow=REQUEST_GENERATOR_DEFINITION_RUN,
                context=graphql_context.get_context(),
                parameters={"model": request_model},
            )
            return cls(ok=True, task={"id": workflow.id})

        await graphql_context.active_service.workflow.execute_workflow(
            workflow=REQUEST_GENERATOR_DEFINITION_RUN,
            context=graphql_context.get_context(),
            parameters={"model": request_model},
        )
        return cls(ok=True)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from pydantic import BaseModel

from infrahub.constants.database import EntityType, IndexType  # noqa: TC001

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class IndexInfo(BaseModel):
    name: str
    label: str
    properties: list[str]
    type: IndexType
    entity_type: EntityType


class IndexItem(BaseModel):
    name: str
    label: str
    properties: list[str]
    type: IndexType

    def get_add_query(self) -> str:
        raise NotImplementedError()

    def get_drop_query(self) -> str:
        raise NotImplementedError()


class IndexManagerBase(ABC):
    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db

        self.nodes: list[IndexItem] = []
        self.rels: list[IndexItem] = []
        self.initialized: bool = False

    def init(self, nodes: list[IndexItem], rels: list[IndexItem]) -> None:
        self.nodes = nodes
        self.rels = rels
        self.initialized = True

    @property
    def items(self) -> list[IndexItem]:
        return self.nodes + self.rels

    async def add(self) -> None:
        async with self.db.start_transaction() as dbt:
            for item in self.items:
                await dbt.execute_query(query=item.get_add_query(), params={}, name="index_add")

    async def drop(self) -> None:
        async with self.db.start_transaction() as dbt:
            for item in self.items:
                await dbt.execute_query(query=item.get_drop_query(), params={}, name="index_drop")

    @abstractmethod
    async def list(self) -> list[IndexInfo]:
        pass



from infrahub.events.branch_action import BranchMergedEvent
from infrahub.trigger.models import BuiltinTriggerDefinition, EventTrigger, ExecuteWorkflow
from infrahub.workflows.catalogue import BRANCH_MERGED

TRIGGER_BRANCH_MERGED = BuiltinTriggerDefinition(
    name="branch-merged-trigger",
    trigger=EventTrigger(
        events={BranchMergedEvent.event_name},
    ),
    actions=[
        ExecuteWorkflow(
            workflow=BRANCH_MERGED,
            parameters={
                "source_branch": "{{ event.payload['data']['branch_name'] }}",
                "context": {
                    "__prefect_kind": "json",
                    "value": {"__prefect_kind": "jinja", "template": "{{ event.payload['context'] | tojson }}"},
                },
            },
        ),
    ],
)



from netaddr import IPNetwork, IPSet

from infrahub.core.ipam.constants import IPAddressType, IPNetworkType


def get_available(network: IPNetworkType, addresses: list[IPAddressType], is_pool: bool) -> IPSet:
    pool = IPSet(IPNetwork(str(network)))
    reserved = [IPNetwork(f"{str(address.ip)}/{address.max_prefixlen}") for address in addresses]
    if not is_pool:
        # If the specified network is not a pool we remove the network address and
        # optionally the broadcast address in case of IPv4
        reserved.append(IPNetwork(f"{str(network.network_address)}/{network.max_prefixlen}"))
        if network.version == 4:
            reserved.append(IPNetwork(f"{str(network.broadcast_address)}/{network.max_prefixlen}"))

    return pool - IPSet(reserved)



from __future__ import annotations

from graphene import Enum, Field, Float, List, ObjectType, String
from graphene.types.generic import GenericScalar
from prefect.client.schemas.objects import StateType

from .task_log import TaskLogEdge

TaskState = Enum.from_enum(StateType)


class TaskInfo(ObjectType):
    id = Field(String)


class Task(ObjectType):
    id = String(required=True)
    title = String(required=True)
    conclusion = String(required=True)
    state = TaskState(required=False)
    progress = Float(required=False)
    workflow = String(required=False)
    branch = String(required=False)
    created_at = String(required=True)
    updated_at = String(required=True)
    parameters = GenericScalar(required=False)
    tags = List(String, required=False)
    start_time = String(required=False)


class TaskRelatedNode(ObjectType):
    id = String(required=True)
    kind = String(required=True)


class TaskNode(Task):
    related_node = String(
        required=False,
        deprecation_reason="This field is deprecated and it will be removed in a future release, use related_nodes instead",
    )
    related_node_kind = String(
        required=False,
        deprecation_reason="This field is deprecated and it will be removed in a future release, use related_nodes instead",
    )
    related_nodes = List(TaskRelatedNode)
    logs = Field(TaskLogEdge)


class TaskNodes(ObjectType):
    node = Field(TaskNode)
